"""Conversion utilities for :class:`Records`.

This module contains helper functions for converting between pandas/polars
DataFrames and :class:`Records`, including schema extraction and type conversion.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, List, Optional

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from ..table.schema import ColumnDef


def is_pandas_dataframe(obj: Any) -> bool:
    """Check if object is a pandas :class:`DataFrame` without importing pandas unnecessarily.

    Args:
        obj: Object to check

    Returns:
        True if object is a pandas :class:`DataFrame`, False otherwise
    """
    module = getattr(type(obj), "__module__", "")
    if "pandas" not in module:
        return False
    try:
        import pandas as pd
    except ImportError:
        return False
    return isinstance(obj, pd.DataFrame)


def is_polars_dataframe(obj: Any) -> bool:
    """Check if object is a polars :class:`DataFrame`.

    Args:
        obj: Object to check

    Returns:
        True if object is a polars :class:`DataFrame`, False otherwise
    """
    module = getattr(type(obj), "__module__", "")
    if "polars" not in module:
        return False
    try:
        import polars as pl
    except ImportError:
        return False
    return isinstance(obj, pl.DataFrame)


def is_polars_lazyframe(obj: Any) -> bool:
    """Check if object is a polars LazyFrame.

    Args:
        obj: Object to check

    Returns:
        True if object is a polars LazyFrame, False otherwise
    """
    module = getattr(type(obj), "__module__", "")
    if "polars" not in module:
        return False
    try:
        import polars as pl
    except ImportError:
        return False
    return isinstance(obj, pl.LazyFrame)


def convert_pandas_dtype_to_sql_type(dtype: Any) -> str:
    """Convert pandas dtype to SQL type name.

    Args:
        dtype: pandas dtype object

    Returns:
        SQL type name string
    """
    dtype_str = str(dtype)

    # Handle nullable integer types
    if dtype_str.startswith("Int"):
        return "INTEGER"

    # Handle standard types
    if dtype_str in ("int64", "int32", "int16", "int8", "int"):
        return "INTEGER"
    if dtype_str in ("float64", "float32", "float"):
        return "REAL"
    if dtype_str in ("bool", "boolean"):
        return "INTEGER"  # SQLite uses INTEGER for booleans
    if dtype_str.startswith("datetime"):
        return "TIMESTAMP"
    if dtype_str.startswith("date"):
        return "DATE"
    if dtype_str.startswith("timedelta"):
        return "TEXT"  # No standard SQL type for timedelta
    if dtype_str in ("object", "string", "str"):
        return "TEXT"

    # Default to TEXT for unknown types
    return "TEXT"


def convert_polars_type_to_sql_type(polars_type: Any) -> str:
    """Convert polars type to SQL type name.

    Args:
        polars_type: polars DataType object or string representation

    Returns:
        SQL type name string
    """
    type_str = str(polars_type).lower()

    # Handle polars types
    if "int" in type_str:
        return "INTEGER"
    if "float" in type_str or "f64" in type_str or "f32" in type_str:
        return "REAL"
    if "bool" in type_str:
        return "INTEGER"  # SQLite uses INTEGER for booleans
    if "datetime" in type_str or "timestamp" in type_str:
        return "TIMESTAMP"
    if "date" in type_str:
        return "DATE"
    if "duration" in type_str or "timedelta" in type_str:
        return "TEXT"  # No standard SQL type for duration
    if "str" in type_str or "string" in type_str or "utf8" in type_str:
        return "TEXT"
    if "null" in type_str:
        return "TEXT"  # Nullable type, default to TEXT

    # Default to TEXT for unknown types
    return "TEXT"


def extract_schema_from_pandas_dataframe(df: Any) -> Optional[List["ColumnDef"]]:
    """Extract schema from pandas :class:`DataFrame`.

    Args:
        df: pandas :class:`DataFrame`

    Returns:
        List of ColumnDef objects or None if extraction fails
    """
    try:
        from ..table.schema import ColumnDef

        columns = []
        for col_name, dtype in df.dtypes.items():
            sql_type = convert_pandas_dtype_to_sql_type(dtype)
            # Check if column has any nulls - convert numpy boolean to Python boolean
            nullable = bool(df[col_name].isna().any())
            columns.append(ColumnDef(name=str(col_name), type_name=sql_type, nullable=nullable))
        return columns
    except (AttributeError, KeyError, TypeError, ValueError) as e:
        logger.debug("Schema extraction from pandas DataFrame failed: %s", e)
        return None
    except Exception as e:
        logger.warning(
            "Unexpected error during pandas DataFrame schema extraction: %s", e, exc_info=True
        )
        return None


def extract_schema_from_polars_dataframe(df: Any) -> Optional[List["ColumnDef"]]:
    """Extract schema from polars :class:`DataFrame`.

    Args:
        df: polars :class:`DataFrame`

    Returns:
        List of ColumnDef objects or None if extraction fails
    """
    try:
        from ..table.schema import ColumnDef

        columns = []
        schema = df.schema
        for col_name, polars_type in schema.items():
            sql_type = convert_polars_type_to_sql_type(polars_type)
            # Check if type is nullable - convert to Python boolean
            nullable = bool(
                polars_type.is_nullable() if hasattr(polars_type, "is_nullable") else True
            )
            columns.append(ColumnDef(name=str(col_name), type_name=sql_type, nullable=nullable))
        return columns
    except (AttributeError, KeyError, TypeError, ValueError) as e:
        logger.debug("Schema extraction from polars DataFrame failed: %s", e)
        return None
    except Exception as e:
        logger.warning(
            "Unexpected error during polars DataFrame schema extraction: %s", e, exc_info=True
        )
        return None


def extract_schema_from_polars_lazyframe(lf: Any) -> Optional[List["ColumnDef"]]:
    """Extract schema from polars LazyFrame.

    Args:
        lf: polars LazyFrame

    Returns:
        List of ColumnDef objects or None if extraction fails
    """
    try:
        from ..table.schema import ColumnDef

        columns = []
        # Use collect_schema() to avoid performance warning
        schema = lf.collect_schema() if hasattr(lf, "collect_schema") else lf.schema
        for col_name, polars_type in schema.items():
            sql_type = convert_polars_type_to_sql_type(polars_type)
            # Check if type is nullable - convert to Python boolean
            nullable = bool(
                polars_type.is_nullable() if hasattr(polars_type, "is_nullable") else True
            )
            columns.append(ColumnDef(name=str(col_name), type_name=sql_type, nullable=nullable))
        return columns
    except (AttributeError, KeyError, TypeError, ValueError) as e:
        logger.debug("Schema extraction from polars LazyFrame failed: %s", e)
        return None
    except Exception as e:
        logger.warning(
            "Unexpected error during polars LazyFrame schema extraction: %s", e, exc_info=True
        )
        return None


def convert_dataframe_to_rows(df: Any) -> List[dict[str, object]]:
    """Convert pandas/polars :class:`DataFrame` or polars LazyFrame to list of dictionaries.

    Args:
        df: pandas :class:`DataFrame`, polars :class:`DataFrame`, or polars LazyFrame

    Returns:
        List of row dictionaries

    Raises:
        ValueError: If :class:`DataFrame` type is not supported
    """
    if is_pandas_dataframe(df):
        # Convert pandas DataFrame to list of dicts
        return df.to_dict("records")  # type: ignore[no-any-return]
    elif is_polars_dataframe(df):
        # Convert polars DataFrame to list of dicts
        return df.to_dicts()  # type: ignore[no-any-return]
    elif is_polars_lazyframe(df):
        # Materialize LazyFrame first, then convert
        materialized = df.collect()
        return materialized.to_dicts()  # type: ignore[no-any-return]
    else:
        raise ValueError(f"Unsupported DataFrame type: {type(df)}")


def extract_schema_from_dataframe(df: Any) -> Optional[List["ColumnDef"]]:
    """Extract schema from pandas/polars :class:`DataFrame` or polars LazyFrame.

    Args:
        df: pandas :class:`DataFrame`, polars :class:`DataFrame`, or polars LazyFrame

    Returns:
        List of ColumnDef objects or None if extraction fails
    """
    if is_pandas_dataframe(df):
        return extract_schema_from_pandas_dataframe(df)
    elif is_polars_dataframe(df):
        return extract_schema_from_polars_dataframe(df)
    elif is_polars_lazyframe(df):
        return extract_schema_from_polars_lazyframe(df)
    return None



"""Logical plan node definitions."""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Dict, Optional

from ..expressions.column import Column

if TYPE_CHECKING:
    from ..table.schema import ColumnDef


@dataclass(frozen=True)
class WindowSpec:
    """Window specification for window functions."""

    partition_by: tuple[Column, ...] = ()
    order_by: tuple[Column, ...] = ()
    rows_between: Optional[tuple[Optional[int], Optional[int]]] = None
    range_between: Optional[tuple[Optional[int], Optional[int]]] = None


@dataclass(frozen=True)
class LogicalPlan:
    """Base class for logical operators."""

    def children(self) -> Sequence[LogicalPlan]:
        return ()


@dataclass(frozen=True)
class TableScan(LogicalPlan):
    table: str
    alias: str | None = None


@dataclass(frozen=True)
class FileScan(LogicalPlan):
    """File scan operation for reading data from files.

    Args:
        path: Path to the file
        format: File format ("csv", "json", "jsonl", "parquet", "text")
        schema: Optional explicit schema for the file data
        options: Dictionary of format-specific read options
        column_name: :class:`Column` name for text files (default: "value")
    """

    path: str
    format: str  # "csv", "json", "jsonl", "parquet", "text"
    schema: Optional[Sequence["ColumnDef"]] = None
    options: Dict[str, object] = field(default_factory=dict)
    column_name: Optional[str] = None  # For text files


@dataclass(frozen=True)
class RawSQL(LogicalPlan):
    """Raw SQL query operation.

    Args:
        sql: SQL query string to execute
        params: Optional dictionary of named parameters for parameterized queries
    """

    sql: str
    params: Dict[str, object] | None = None


@dataclass(frozen=True)
class Project(LogicalPlan):
    child: LogicalPlan
    projections: tuple[Column, ...]

    def children(self) -> Sequence[LogicalPlan]:
        return (self.child,)


@dataclass(frozen=True)
class Filter(LogicalPlan):
    child: LogicalPlan
    predicate: Column

    def children(self) -> Sequence[LogicalPlan]:
        return (self.child,)


@dataclass(frozen=True)
class Limit(LogicalPlan):
    child: LogicalPlan
    count: int
    offset: int = 0

    def children(self) -> Sequence[LogicalPlan]:
        return (self.child,)


@dataclass(frozen=True)
class Sample(LogicalPlan):
    """Sample rows from a :class:`DataFrame`.

    Args:
        child: The logical plan to sample from
        fraction: Fraction of rows to sample (0.0 to 1.0)
        seed: Optional random seed for reproducible sampling
    """

    child: LogicalPlan
    fraction: float
    seed: Optional[int] = None

    def children(self) -> Sequence[LogicalPlan]:
        return (self.child,)


@dataclass(frozen=True)
class SortOrder:
    expression: Column
    descending: bool = False


@dataclass(frozen=True)
class Sort(LogicalPlan):
    child: LogicalPlan
    orders: tuple[SortOrder, ...]

    def children(self) -> Sequence[LogicalPlan]:
        return (self.child,)


@dataclass(frozen=True)
class Aggregate(LogicalPlan):
    child: LogicalPlan
    grouping: tuple[Column, ...]
    aggregates: tuple[Column, ...]

    def children(self) -> Sequence[LogicalPlan]:
        return (self.child,)


@dataclass(frozen=True)
class Join(LogicalPlan):
    left: LogicalPlan
    right: LogicalPlan
    how: str
    on: tuple[tuple[str, str], ...] | None = None
    condition: Column | None = None
    lateral: bool = False  # True for LATERAL join (PostgreSQL, MySQL 8.0+)
    hints: tuple[str, ...] | None = None  # Join hints (e.g., "USE_INDEX", "FORCE_INDEX")

    def children(self) -> Sequence[LogicalPlan]:
        return (self.left, self.right)


@dataclass(frozen=True)
class Distinct(LogicalPlan):
    child: LogicalPlan

    def children(self) -> Sequence[LogicalPlan]:
        return (self.child,)


@dataclass(frozen=True)
class Union(LogicalPlan):
    left: LogicalPlan
    right: LogicalPlan
    distinct: bool = True  # True for UNION, False for UNION ALL

    def children(self) -> Sequence[LogicalPlan]:
        return (self.left, self.right)


@dataclass(frozen=True)
class Intersect(LogicalPlan):
    left: LogicalPlan
    right: LogicalPlan
    distinct: bool = True  # True for INTERSECT, False for INTERSECT ALL

    def children(self) -> Sequence[LogicalPlan]:
        return (self.left, self.right)


@dataclass(frozen=True)
class Except(LogicalPlan):
    left: LogicalPlan
    right: LogicalPlan
    distinct: bool = True  # True for EXCEPT, False for EXCEPT ALL

    def children(self) -> Sequence[LogicalPlan]:
        return (self.left, self.right)


@dataclass(frozen=True)
class CTE(LogicalPlan):
    """Common Table Expression (CTE) - a named subquery that can be referenced later."""

    name: str
    child: LogicalPlan

    def children(self) -> Sequence[LogicalPlan]:
        return (self.child,)


@dataclass(frozen=True)
class RecursiveCTE(LogicalPlan):
    """Recursive Common Table Expression (WITH RECURSIVE) - for recursive queries."""

    name: str
    initial: LogicalPlan  # Initial/seed query
    recursive: LogicalPlan  # Recursive part that references the CTE
    union_all: bool = False  # True for UNION ALL, False for UNION

    def children(self) -> Sequence[LogicalPlan]:
        return (self.initial, self.recursive)


@dataclass(frozen=True)
class SemiJoin(LogicalPlan):
    """Semi-join: returns rows from left where a matching row exists in right (EXISTS subquery)."""

    left: LogicalPlan
    right: LogicalPlan
    on: tuple[tuple[str, str], ...] | None = None
    condition: Column | None = None

    def children(self) -> Sequence[LogicalPlan]:
        return (self.left, self.right)


@dataclass(frozen=True)
class AntiJoin(LogicalPlan):
    """Anti-join: returns rows from left where no matching row exists in right (NOT EXISTS subquery)."""

    left: LogicalPlan
    right: LogicalPlan
    on: tuple[tuple[str, str], ...] | None = None
    condition: Column | None = None

    def children(self) -> Sequence[LogicalPlan]:
        return (self.left, self.right)


@dataclass(frozen=True)
class Pivot(LogicalPlan):
    """Pivot operation for data reshaping.

    Args:
        child: The logical plan to pivot
        pivot_column: :class:`Column` to pivot on (becomes column headers)
        value_column: :class:`Column` containing values to aggregate
        agg_func: Aggregation function name (e.g., "sum", "avg", "count")
        pivot_values: Optional list of specific values to pivot (if None, uses all distinct values)
    """

    child: LogicalPlan
    pivot_column: str
    value_column: str
    agg_func: str
    pivot_values: tuple[str, ...] | None = None

    def children(self) -> Sequence[LogicalPlan]:
        return (self.child,)


@dataclass(frozen=True)
class GroupedPivot(LogicalPlan):
    """Grouped pivot operation that combines GROUP BY with pivot.

    Args:
        child: The logical plan to pivot
        grouping: Columns to group by
        pivot_column: :class:`Column` to pivot on (becomes column headers)
        value_column: :class:`Column` containing values to aggregate
        agg_func: Aggregation function name (e.g., "sum", "avg", "count")
        pivot_values: Optional list of specific values to pivot (if None, must be provided or discovered)
    """

    child: LogicalPlan
    grouping: tuple[Column, ...]
    pivot_column: str
    value_column: str
    agg_func: str
    pivot_values: tuple[str, ...] | None = None

    def children(self) -> Sequence[LogicalPlan]:
        return (self.child,)


@dataclass(frozen=True)
class Explode(LogicalPlan):
    """Explode: expands array/JSON column into multiple rows (one row per element)."""

    child: LogicalPlan
    column: Column
    alias: str = "value"  # Alias for the exploded value column

    def children(self) -> Sequence[LogicalPlan]:
        return (self.child,)



"""Pytest integration package for Moltres."""

from __future__ import annotations

__all__ = [
    "moltres_db",
    "moltres_async_db",
    "test_data",
    "create_test_df",
    "assert_dataframe_equal",
    "assert_schema_equal",
    "assert_query_results",
    "_test_data_fixture",
    "query_logger",
]

try:
    from .fixtures import (
        _test_data_fixture,
        assert_dataframe_equal,
        assert_query_results,
        assert_schema_equal,
        create_test_df,
        moltres_async_db,
        moltres_db,
        test_data,
    )
    from .plugin import query_logger
except ImportError:
    # Pytest not available or import failed
    moltres_db = None  # type: ignore[assignment]
    moltres_async_db = None  # type: ignore[assignment]
    test_data = None  # type: ignore[assignment]
    _test_data_fixture = None  # type: ignore[assignment]
    create_test_df = None  # type: ignore[assignment]
    assert_dataframe_equal = None  # type: ignore[assignment]
    assert_schema_equal = None  # type: ignore[assignment]
    assert_query_results = None  # type: ignore[assignment]
    query_logger = None  # type: ignore[assignment]



"""dbt integration for Moltres.

This module provides integration with dbt (data build tool) for using
Moltres DataFrames in dbt Python models.

Key features:
- Get Moltres :class:`Database` instances from dbt connections
- Reference dbt models and sources as Moltres DataFrames
- Helper functions for common dbt patterns
"""

from __future__ import annotations

from typing import Any

__all__ = [
    "get_moltres_connection",
    "moltres_ref",
    "moltres_source",
    "moltres_var",
]

# Graceful degradation
try:
    import dbt

    DBT_AVAILABLE = True
except ImportError:
    DBT_AVAILABLE = False
    # Create stubs for type checking
    dbt: Any = None  # type: ignore[no-redef]

from .adapter import get_moltres_connection
from .helpers import moltres_ref, moltres_source, moltres_var



"""Synchronous SQLAlchemy integration helpers for Moltres.

This module provides helper functions for integrating Moltres with existing
SQLAlchemy projects, allowing you to use Moltres DataFrames with existing
SQLAlchemy connections, sessions, and infrastructure.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type

if TYPE_CHECKING:
    from sqlalchemy.engine import Connection
    from sqlalchemy.orm import Session
    from sqlalchemy.sql import Select
    from ...dataframe.core.dataframe import DataFrame
    from ...table.table import Database


def execute_with_connection(df: "DataFrame", connection: "Connection") -> List[Dict[str, Any]]:
    """Execute a Moltres :class:`DataFrame` using a provided SQLAlchemy Connection.

    This allows you to execute Moltres queries within an existing SQLAlchemy
    transaction or connection context.

    Args:
        df: Moltres :class:`DataFrame` to execute
        connection: SQLAlchemy Connection to use for execution

    Returns:
        List of dictionaries representing rows

    Example:
        >>> from sqlalchemy import create_engine
        >>> from moltres import connect, col
        >>> from moltres.integrations.sqlalchemy import execute_with_connection
        >>> engine = create_engine("sqlite:///:memory:")
        >>> db = connect(engine=engine)
        >>> df = db.table("users").select().where(col("id") > 1)
        >>> with engine.connect() as conn:
        ...     results = execute_with_connection(df, conn)
    """
    # Convert DataFrame to SQLAlchemy statement
    stmt = df.to_sqlalchemy()

    # Execute using the provided connection
    result = connection.execute(stmt)
    rows = result.fetchall()
    columns = list(result.keys())

    # Format as list of dicts
    return [dict(zip(columns, row)) for row in rows]


def execute_with_session(df: "DataFrame", session: "Session") -> List[Dict[str, Any]]:
    """Execute a Moltres :class:`DataFrame` using a SQLAlchemy ORM Session.

    This allows you to execute Moltres queries within an existing SQLAlchemy
    ORM session context.

    Args:
        df: Moltres :class:`DataFrame` to execute
        session: SQLAlchemy ORM Session to use for execution

    Returns:
        List of dictionaries representing rows

    Example:
        >>> from sqlalchemy import create_engine
        >>> from sqlalchemy.orm import sessionmaker
        >>> from moltres import connect, col
        >>> from moltres.integrations.sqlalchemy import execute_with_session
        >>> engine = create_engine("sqlite:///:memory:")
        >>> Session = sessionmaker(bind=engine)
        >>> db = connect(engine=engine)
        >>> df = db.table("users").select().where(col("id") > 1)
        >>> with Session() as session:
        ...     results = execute_with_session(df, session)
    """
    # Get connection from session
    connection = session.connection()
    return execute_with_connection(df, connection)


def to_sqlalchemy_select(df: "DataFrame", dialect: Optional[str] = None) -> "Select":
    """Convert a Moltres :class:`DataFrame` to a SQLAlchemy Select statement.

    This is a convenience function that wraps :class:`DataFrame`.to_sqlalchemy().

    Args:
        df: Moltres :class:`DataFrame` to convert
        dialect: Optional SQL dialect name. If not provided, uses the dialect
                from the :class:`DataFrame`'s attached :class:`Database`, or defaults to "ansi"

    Returns:
        SQLAlchemy Select statement

    Example:
        >>> from moltres import connect, col
        >>> from moltres.integrations.sqlalchemy import to_sqlalchemy_select
        >>> db = connect("sqlite:///:memory:")
        >>> df = db.table("users").select().where(col("id") > 1)
        >>> stmt = to_sqlalchemy_select(df)
        >>> # Now use stmt with any SQLAlchemy connection
    """
    return df.to_sqlalchemy(dialect=dialect)


def from_sqlalchemy_select(
    select_stmt: "Select", database: Optional["Database"] = None
) -> "DataFrame":
    """Create a Moltres :class:`DataFrame` from a SQLAlchemy Select statement.

    This is a convenience function that wraps :class:`DataFrame`.from_sqlalchemy().

    Args:
        select_stmt: SQLAlchemy Select statement to convert
        database: Optional :class:`Database` instance to attach to the :class:`DataFrame`

    Returns:
        Moltres :class:`DataFrame` that can be further chained with Moltres operations

    Example:
        >>> from sqlalchemy import create_engine, select, table, column
        >>> from moltres.integrations.sqlalchemy import from_sqlalchemy_select
        >>> engine = create_engine("sqlite:///:memory:")
        >>> users = table("users", column("id"), column("name"))
        >>> sa_stmt = select(users.c.id, users.c.name).where(users.c.id > 1)
        >>> df = from_sqlalchemy_select(sa_stmt)
        >>> # Can now chain Moltres operations
    """
    from ...dataframe.core.dataframe import DataFrame

    return DataFrame.from_sqlalchemy(select_stmt, database=database)


def with_sqlmodel(df: "DataFrame", model: Type[Any]) -> "DataFrame":
    """Attach a SQLModel or Pydantic model to a :class:`DataFrame`.

    This is a convenience function that wraps :class:`DataFrame`.with_model().

    Args:
        df: Moltres :class:`DataFrame`
        model: SQLModel or Pydantic model class to attach

    Returns:
        :class:`DataFrame` with the model attached

    Example:
        >>> from sqlmodel import SQLModel, Field
        >>> from moltres import connect
        >>> from moltres.integrations.sqlalchemy import with_sqlmodel
        >>> class User(SQLModel, table=True):
        ...     id: int = Field(primary_key=True)
        ...     name: str
        >>> db = connect("sqlite:///:memory:")
        >>> df = db.table("users").select()
        >>> df_with_model = with_sqlmodel(df, User)
        >>> results = df_with_model.collect()  # Returns list of User instances

        >>> from pydantic import BaseModel
        >>> class UserData(BaseModel):
        ...     id: int
        ...     name: str
        >>> df_with_pydantic = with_sqlmodel(df, UserData)
        >>> results = df_with_pydantic.collect()  # Returns list of UserData instances
    """
    return df.with_model(model)


def execute_with_connection_model(
    df: "DataFrame", connection: "Connection", model: Type[Any]
) -> List[Any]:
    """Execute a Moltres :class:`DataFrame` using a provided SQLAlchemy Connection and return SQLModel instances.

    Args:
        df: Moltres :class:`DataFrame` to execute
        connection: SQLAlchemy Connection to use for execution
        model: SQLModel model class to instantiate results as

    Returns:
        List of SQLModel instances

    Example:
        >>> from sqlmodel import SQLModel, Field
        >>> from sqlalchemy import create_engine
        >>> from moltres import connect, col
        >>> from moltres.integrations.sqlalchemy import execute_with_connection_model
        >>> class User(SQLModel, table=True):
        ...     id: int = Field(primary_key=True)
        ...     name: str
        >>> engine = create_engine("sqlite:///:memory:")
        >>> db = connect(engine=engine)
        >>> df = db.table("users").select().where(col("id") > 1)
        >>> with engine.connect() as conn:
        ...     results = execute_with_connection_model(df, conn, User)
    """
    df_with_model = df.with_model(model)
    # Execute using the existing function and convert to models
    dict_rows = execute_with_connection(df_with_model, connection)

    # Convert to SQLModel instances
    from ...utils.sqlmodel_integration import rows_to_sqlmodels

    return rows_to_sqlmodels(dict_rows, model)


def execute_with_session_model(df: "DataFrame", session: "Session", model: Type[Any]) -> List[Any]:
    """Execute a Moltres :class:`DataFrame` using a SQLAlchemy ORM Session and return SQLModel instances.

    Args:
        df: Moltres :class:`DataFrame` to execute
        session: SQLAlchemy ORM Session to use for execution
        model: SQLModel model class to instantiate results as

    Returns:
        List of SQLModel instances

    Example:
        >>> from sqlmodel import SQLModel, Field
        >>> from sqlalchemy import create_engine
        >>> from sqlalchemy.orm import sessionmaker
        >>> from moltres import connect, col
        >>> from moltres.integrations.sqlalchemy import execute_with_session_model
        >>> class User(SQLModel, table=True):
        ...     id: int = Field(primary_key=True)
        ...     name: str
        >>> engine = create_engine("sqlite:///:memory:")
        >>> Session = sessionmaker(bind=engine)
        >>> db = connect(engine=engine)
        >>> df = db.table("users").select().where(col("id") > 1)
        >>> with Session() as session:
        ...     results = execute_with_session_model(df, session, User)
    """
    # Get connection from session
    connection = session.connection()
    return execute_with_connection_model(df, connection, model)


__all__ = [
    "execute_with_connection",
    "execute_with_session",
    "to_sqlalchemy_select",
    "from_sqlalchemy_select",
    "with_sqlmodel",
    "execute_with_connection_model",
    "execute_with_session_model",
]



"""Aggregation functions for :class:`DataFrame` operations."""

from __future__ import annotations

from typing import Union

from ..column import Column, ColumnLike, ensure_column


def _aggregate(op: str, column: ColumnLike) -> Column:
    """Internal helper for creating aggregate expressions."""
    return Column(op=op, args=(ensure_column(column),))


def sum(column: ColumnLike) -> Column:  # noqa: A001 - mirrored PySpark API
    """Compute the sum of a column.

    Args:
        column: :class:`Column` expression or literal value

    Returns:
        :class:`Column` expression for the sum aggregate

    Example:
        >>> from moltres import connect, col
        >>> from moltres.expressions import functions as F
        >>> from moltres.table.schema import column
        >>> db = connect("sqlite:///:memory:")
        >>> _ = db.create_table("sales", [column("category", "TEXT"), column("amount", "REAL"), column("status", "TEXT")]).collect()  # doctest: +ELLIPSIS
        >>> from moltres.io.records import :class:`Records`
        >>> _ = :class:`Records`(_data=[{"category": "A", "amount": 100.0, "status": "active"}, {"category": "A", "amount": 200.0, "status": "completed"}], _database=db).insert_into("sales")
            >>> # Sum aggregation
            >>> df = db.table("sales").select().group_by("category").agg(F.sum(col("amount")).alias("total"))
        >>> results = df.collect()
        >>> results[0]["total"]
        300.0
            >>> # With FILTER clause for conditional aggregation
            >>> df2 = db.table("sales").select().group_by("category").agg(F.sum(col("amount")).filter(col("status") == "active").alias("active_total"))
        >>> results2 = df2.collect()
        >>> results2[0]["active_total"]
        100.0
        >>> db.close()
    """
    return _aggregate("agg_sum", column)


def avg(column: ColumnLike) -> Column:
    """Compute the average of a column.

    Args:
        column: :class:`Column` expression or literal value

    Returns:
        :class:`Column` expression for the average aggregate

    Example:
        >>> from moltres import connect, col
        >>> from moltres.expressions import functions as F
        >>> from moltres.table.schema import column
        >>> db = connect("sqlite:///:memory:")
        >>> _ = db.create_table("products", [column("category", "TEXT"), column("price", "REAL"), column("active", "INTEGER")]).collect()  # doctest: +ELLIPSIS
        >>> from moltres.io.records import :class:`Records`
        >>> _ = :class:`Records`(_data=[{"category": "A", "price": 10.0, "active": 1}, {"category": "A", "price": 20.0, "active": 1}], _database=db).insert_into("products")
            >>> # Average aggregation
            >>> df = db.table("products").select().group_by("category").agg(F.avg(col("price")).alias("avg_price"))
        >>> results = df.collect()
        >>> results[0]["avg_price"]
        15.0
        >>> db.close()
    """
    return _aggregate("agg_avg", column)


def min(column: ColumnLike) -> Column:  # noqa: A001 - mirrored PySpark API
    """Compute the minimum value of a column.

    Args:
        column: :class:`Column` expression or literal value

    Returns:
        :class:`Column` expression for the minimum aggregate

    Example:
        >>> from moltres import connect, col
        >>> from moltres.expressions import functions as F
        >>> from moltres.table.schema import column
        >>> db = connect("sqlite:///:memory:")
        >>> _ = db.create_table("products", [column("category", "TEXT"), column("price", "REAL")]).collect()  # doctest: +ELLIPSIS
        >>> from moltres.io.records import :class:`Records`
        >>> _ = :class:`Records`(_data=[{"category": "A", "price": 10.0}, {"category": "A", "price": 20.0}], _database=db).insert_into("products")
            >>> # Minimum aggregation
            >>> df = db.table("products").select().group_by("category").agg(F.min(col("price")).alias("min_price"))
        >>> results = df.collect()
        >>> results[0]["min_price"]
        10.0
        >>> db.close()
    """
    return _aggregate("agg_min", column)


def max(column: ColumnLike) -> Column:  # noqa: A001 - mirrored PySpark API
    """Compute the maximum value of a column.

    Args:
        column: :class:`Column` expression or literal value

    Returns:
        :class:`Column` expression for the maximum aggregate

    Example:
        >>> from moltres import connect, col
        >>> from moltres.expressions import functions as F
        >>> from moltres.table.schema import column
        >>> db = connect("sqlite:///:memory:")
        >>> _ = db.create_table("products", [column("category", "TEXT"), column("price", "REAL")]).collect()  # doctest: +ELLIPSIS
        >>> from moltres.io.records import :class:`Records`
        >>> _ = :class:`Records`(_data=[{"category": "A", "price": 10.0}, {"category": "A", "price": 20.0}], _database=db).insert_into("products")
            >>> # Maximum aggregation
            >>> df = db.table("products").select().group_by("category").agg(F.max(col("price")).alias("max_price"))
        >>> results = df.collect()
        >>> results[0]["max_price"]
        20.0
        >>> db.close()
    """
    return _aggregate("agg_max", column)


def count(column: Union[ColumnLike, str] = "*") -> Column:
    """Count the number of rows or non-null values.

    Args:
        column: :class:`Column` expression, literal value, or "*" for counting all rows

    Returns:
        :class:`Column` expression for the count aggregate

    Example:
        >>> from moltres import connect, col
        >>> from moltres.expressions import functions as F
        >>> from moltres.table.schema import column
        >>> db = connect("sqlite:///:memory:")
        >>> _ = db.create_table("products", [column("category", "TEXT"), column("id", "INTEGER")]).collect()  # doctest: +ELLIPSIS
        >>> from moltres.io.records import :class:`Records`
        >>> _ = :class:`Records`(_data=[{"category": "A", "id": 1}, {"category": "A", "id": 2}], _database=db).insert_into("products")
        >>> # Count all rows
        >>> df = db.table("products").select().group_by("category").agg(F.count("*").alias("count"))
        >>> results = df.collect()
        >>> results[0]["count"]
        2
        >>> # Count non-null values in a column
        >>> df2 = db.table("products").select().group_by("category").agg(F.count(col("id")).alias("id_count"))
        >>> results2 = df2.collect()
        >>> results2[0]["id_count"]
        2
        >>> db.close()
    """
    if isinstance(column, str) and column == "*":
        return Column(op="agg_count_star", args=())
    return _aggregate("agg_count", column)


def count_distinct(*columns: ColumnLike) -> Column:
    """Count distinct values in one or more columns.

    Args:
        *columns: One or more column expressions

    Returns:
        :class:`Column` expression for the count distinct aggregate

    Example:
        >>> from moltres import connect, col
        >>> from moltres.expressions import functions as F
        >>> from moltres.table.schema import column
        >>> db = connect("sqlite:///:memory:")
        >>> _ = db.create_table("orders", [column("category", "TEXT"), column("user_id", "INTEGER")]).collect()  # doctest: +ELLIPSIS
        >>> from moltres.io.records import :class:`Records`
        >>> _ = :class:`Records`(_data=[{"category": "A", "user_id": 1}, {"category": "A", "user_id": 2}, {"category": "A", "user_id": 1}], _database=db).insert_into("orders")
        >>> df = db.table("orders").select().group_by("category").agg(F.count_distinct(col("user_id")).alias("distinct_users"))
        >>> results = df.collect()
        >>> results[0]["distinct_users"]
        2
        >>> db.close()
    """
    if not columns:
        raise ValueError("count_distinct requires at least one column")
    exprs = tuple(ensure_column(column) for column in columns)
    return Column(op="agg_count_distinct", args=exprs)


def stddev(column: ColumnLike) -> Column:
    """Compute the standard deviation of a column.

    Args:
        column: :class:`Column` expression or literal value

    Returns:
        :class:`Column` expression for the standard deviation aggregate

    Example:
        >>> # Note: stddev() requires database-specific support (PostgreSQL/MySQL) (DuckDB/PostgreSQL/MySQL)
        >>> # SQLite does not have stddev function
        >>> from moltres import connect, col
        >>> from moltres.expressions import functions as F
        >>> from moltres.table.schema import column
        >>> db = connect("duckdb:///:memory:")
        >>> _ = db.create_table("sales", [column("category", "TEXT"), column("amount", "REAL")]).collect()
        >>> from moltres.io.records import :class:`Records`
        >>> _ = :class:`Records`(_data=[{"category": "A", "amount": 100.0}, {"category": "A", "amount": 200.0}], _database=db).insert_into("sales")
        >>> df = db.table("sales").select().group_by("category").agg(F.stddev(col("amount")).alias("std"))
        >>> results = df.collect()
        >>> results[0]["std"] > 0
        True
        >>> db.close()
    """
    return _aggregate("agg_stddev", column)


def variance(column: ColumnLike) -> Column:
    """Compute the variance of a column.

    Args:
        column: :class:`Column` expression or literal value

    Returns:
        :class:`Column` expression for the variance aggregate

    Example:
        >>> # Note: variance() requires database-specific support (PostgreSQL/MySQL) (DuckDB/PostgreSQL/MySQL)
        >>> # SQLite does not have variance function
        >>> from moltres import connect, col
        >>> from moltres.expressions import functions as F
        >>> from moltres.table.schema import column
        >>> db = connect("duckdb:///:memory:")
        >>> _ = db.create_table("sales", [column("category", "TEXT"), column("amount", "REAL")]).collect()
        >>> from moltres.io.records import :class:`Records`
        >>> _ = :class:`Records`(_data=[{"category": "A", "amount": 100.0}, {"category": "A", "amount": 200.0}], _database=db).insert_into("sales")
        >>> df = db.table("sales").select().group_by("category").agg(F.variance(col("amount")).alias("var"))
        >>> results = df.collect()
        >>> results[0]["var"] > 0
        True
        >>> db.close()
    """
    return _aggregate("agg_variance", column)


def corr(column1: ColumnLike, column2: ColumnLike) -> Column:
    """Compute the correlation coefficient between two columns.

    Args:
        column1: First column expression
        column2: Second column expression

    Returns:
        :class:`Column` expression for the correlation aggregate

    Example:
        >>> # Note: corr() requires database-specific support (PostgreSQL/MySQL) (DuckDB/PostgreSQL/MySQL)
        >>> # SQLite does not have corr function
        >>> from moltres import connect, col
        >>> from moltres.expressions import functions as F
        >>> from moltres.table.schema import column
        >>> db = connect("duckdb:///:memory:")
        >>> _ = db.create_table("data", [column("x", "REAL"), column("y", "REAL")]).collect()
        >>> from moltres.io.records import :class:`Records`
        >>> _ = :class:`Records`(_data=[{"x": 1.0, "y": 2.0}, {"x": 2.0, "y": 4.0}], _database=db).insert_into("data")
        >>> # For global aggregation, select the aggregation directly
        >>> df = db.table("data").select(F.corr(col("x"), col("y")).alias("correlation"))
        >>> results = df.collect()
        >>> -1.0 <= results[0]["correlation"] <= 1.0
        True
        >>> db.close()
    """
    return Column(op="agg_corr", args=(ensure_column(column1), ensure_column(column2)))


def covar(column1: ColumnLike, column2: ColumnLike) -> Column:
    """Compute the covariance between two columns.

    Args:
        column1: First column expression
        column2: Second column expression

    Returns:
        :class:`Column` expression for the covariance aggregate

    Example:
        >>> # Note: covar() requires database-specific support (PostgreSQL/MySQL) (DuckDB/PostgreSQL/MySQL)
        >>> # SQLite does not have covar function
        >>> from moltres import connect, col
        >>> from moltres.expressions import functions as F
        >>> from moltres.table.schema import column
        >>> db = connect("duckdb:///:memory:")
        >>> _ = db.create_table("data", [column("x", "REAL"), column("y", "REAL")]).collect()
        >>> from moltres.io.records import :class:`Records`
        >>> _ = :class:`Records`(_data=[{"x": 1.0, "y": 2.0}, {"x": 2.0, "y": 4.0}], _database=db).insert_into("data")
        >>> # For global aggregation, select the aggregation directly
        >>> df = db.table("data").select(F.covar(col("x"), col("y")).alias("covariance"))
        >>> results = df.collect()
        >>> isinstance(results[0]["covariance"], (int, float))
        True
        >>> db.close()
    """
    return Column(op="agg_covar", args=(ensure_column(column1), ensure_column(column2)))


def percentile_cont(column: ColumnLike, fraction: float) -> Column:
    """Compute the continuous percentile (interpolated) of a column.

    Args:
        column: :class:`Column` expression to compute percentile for
        fraction: Percentile fraction (0.0 to 1.0, e.g., 0.5 for median)

    Returns:
        :class:`Column` expression for percentile_cont aggregate

    Example:
        >>> # Note: percentile_cont() requires database-specific support (PostgreSQL/MySQL) (DuckDB/PostgreSQL/MySQL)
        >>> # SQLite does not have percentile_cont function
        >>> from moltres import connect, col
        >>> from moltres.expressions import functions as F
        >>> from moltres.table.schema import column
        >>> db = connect("duckdb:///:memory:")
        >>> _ = db.create_table("sales", [column("category", "TEXT"), column("price", "REAL")]).collect()
        >>> from moltres.io.records import :class:`Records`
        >>> _ = :class:`Records`(_data=[{"category": "A", "price": 100.0}, {"category": "A", "price": 200.0}], _database=db).insert_into("sales")
        >>> df = db.table("sales").select().group_by("category").agg(F.percentile_cont(col("price"), 0.5).alias("median_price"))
        >>> results = df.collect()
        >>> 100.0 <= results[0]["median_price"] <= 200.0
        True
        >>> db.close()
    """
    if not 0.0 <= fraction <= 1.0:
        raise ValueError("fraction must be between 0.0 and 1.0")
    return Column(op="agg_percentile_cont", args=(ensure_column(column), fraction))


def percentile_disc(column: ColumnLike, fraction: float) -> Column:
    """Compute the discrete percentile (actual value) of a column.

    Args:
        column: :class:`Column` expression to compute percentile for
        fraction: Percentile fraction (0.0 to 1.0, e.g., 0.5 for median)

    Returns:
        :class:`Column` expression for percentile_disc aggregate

    Example:
        >>> # Note: percentile_disc() requires database-specific support (PostgreSQL/MySQL) (DuckDB/PostgreSQL/MySQL)
        >>> # SQLite does not have percentile_disc function
        >>> from moltres import connect, col
        >>> from moltres.expressions import functions as F
        >>> from moltres.table.schema import column
        >>> db = connect("duckdb:///:memory:")
        >>> _ = db.create_table("sales", [column("category", "TEXT"), column("price", "REAL")]).collect()
        >>> from moltres.io.records import :class:`Records`
        >>> _ = :class:`Records`(_data=[{"category": "A", "price": 100.0}, {"category": "A", "price": 200.0}], _database=db).insert_into("sales")
        >>> df = db.table("sales").select().group_by("category").agg(F.percentile_disc(col("price"), 0.9).alias("p90_price"))
        >>> results = df.collect()
        >>> results[0]["p90_price"] in [100.0, 200.0]
        True
        >>> db.close()
    """
    if not 0.0 <= fraction <= 1.0:
        raise ValueError("fraction must be between 0.0 and 1.0")
    return Column(op="agg_percentile_disc", args=(ensure_column(column), fraction))


def collect_list(column: ColumnLike) -> Column:
    """Collect values from a column into an array (aggregate function).

    Args:
        column: :class:`Column` expression to collect

    Returns:
        :class:`Column` expression for collect_list aggregate

    Example:
        >>> # Note: collect_list() requires database-specific array support (PostgreSQL/MySQL)
        >>> # SQLite does not support arrays natively
        >>> from moltres import connect, col
        >>> from moltres.expressions import functions as F
        >>> from moltres.table.schema import column
        >>> db = connect("duckdb:///:memory:")
        >>> _ = db.create_table("items", [column("category", "TEXT"), column("item", "TEXT")]).collect()
        >>> from moltres.io.records import :class:`Records`
        >>> _ = :class:`Records`(_data=[{"category": "A", "item": "x"}, {"category": "A", "item": "y"}], _database=db).insert_into("items")
        >>> df = db.table("items").select().group_by("category").agg(F.collect_list(col("item")).alias("items_list"))
        >>> results = df.collect()
        >>> len(results[0]["items_list"])
        2
        >>> db.close()
    """
    return _aggregate("agg_collect_list", column)


def collect_set(column: ColumnLike) -> Column:
    """Collect distinct values from a column into an array (aggregate function).

    Args:
        column: :class:`Column` expression to collect

    Returns:
        :class:`Column` expression for collect_set aggregate

    Example:
        >>> # Note: collect_set() requires database-specific array support (PostgreSQL/MySQL)
        >>> # SQLite does not support arrays natively
        >>> from moltres import connect, col
        >>> from moltres.expressions import functions as F
        >>> from moltres.table.schema import column
        >>> db = connect("duckdb:///:memory:")
        >>> _ = db.create_table("items", [column("category", "TEXT"), column("item", "TEXT")]).collect()
        >>> from moltres.io.records import :class:`Records`
        >>> _ = :class:`Records`(_data=[{"category": "A", "item": "x"}, {"category": "A", "item": "x"}], _database=db).insert_into("items")
        >>> df = db.table("items").select().group_by("category").agg(F.collect_set(col("item")).alias("items_set"))
        >>> results = df.collect()
        >>> len(results[0]["items_set"])
        1
        >>> db.close()
    """
    return _aggregate("agg_collect_set", column)



"""FastAPI integration package for Moltres."""

from __future__ import annotations

__all__ = [
    "register_exception_handlers",
    "get_db",
    "get_async_db",
    "create_db_dependency",
    "create_async_db_dependency",
    "handle_moltres_errors",
    "FASTAPI_AVAILABLE",
]

try:
    from .core import (
        FASTAPI_AVAILABLE,
        create_async_db_dependency,
        create_db_dependency,
        get_async_db,
        get_db,
        handle_moltres_errors,
        register_exception_handlers,
    )
except ImportError:
    # FastAPI not available or import failed
    FASTAPI_AVAILABLE = False
    register_exception_handlers = None  # type: ignore[assignment]
    get_db = None  # type: ignore[assignment]
    get_async_db = None  # type: ignore[assignment]
    create_db_dependency = None  # type: ignore[assignment]
    create_async_db_dependency = None  # type: ignore[assignment]
    handle_moltres_errors = None  # type: ignore[assignment]



"""Async Pandas-style indexers for Moltres DataFrames."""

from __future__ import annotations

from typing import Any

from ...expressions.column import Column
from ..interfaces.async_pandas_dataframe import AsyncPandasDataFrame


class _AsyncLocIndexer:
    """Indexer for pandas-style loc accessor (async)."""

    _df: AsyncPandasDataFrame

    def __init__(self, df: AsyncPandasDataFrame):
        self._df = df

    def __getitem__(self, key: Any) -> AsyncPandasDataFrame:
        """Access rows and columns using loc.

        Supports:
        - df.loc[df['age'] > 25] - Row filtering
        - df.loc[:, ['col1', 'col2']] - :class:`Column` selection
        - df.loc[df['age'] > 25, 'col1'] - Combined filter and select
        """
        # Handle different key types
        if isinstance(key, tuple) and len(key) == 2:
            # Two-dimensional indexing: df.loc[rows, cols]
            row_key, col_key = key
            result_df = self._df._df

            # Apply row filter if not ':' or Ellipsis
            # Need to check type first to avoid boolean evaluation of Column
            if isinstance(row_key, Column):
                # Boolean condition
                result_df = result_df.where(row_key)
            elif row_key is not Ellipsis:
                # Check if it's slice(None) without triggering comparison
                if not (
                    isinstance(row_key, slice)
                    and row_key.start is None
                    and row_key.stop is None
                    and row_key.step is None
                ):
                    raise NotImplementedError(
                        "loc row indexing only supports boolean conditions or :"
                    )

            # Apply column selection
            if col_key is Ellipsis or (
                isinstance(col_key, slice)
                and col_key.start is None
                and col_key.stop is None
                and col_key.step is None
            ):
                # Select all columns (no change)
                pass
            elif isinstance(col_key, (list, tuple)):
                # Select specific columns
                str_columns = [c for c in col_key if isinstance(c, str)]
                if str_columns:
                    self._df._validate_columns_exist(str_columns, "loc column selection")
                from ...expressions.column import col

                columns = [col(c) if isinstance(c, str) else c for c in col_key]
                result_df = result_df.select(*columns)
            elif isinstance(col_key, str):
                # Single column
                self._df._validate_columns_exist([col_key], "loc column selection")
                from ...expressions.column import col

                result_df = result_df.select(col(col_key))
            else:
                raise TypeError(f"Invalid column key type for loc: {type(col_key)}")

            return self._df._with_dataframe(result_df)
        elif isinstance(key, Column):
            # Single boolean condition - filter rows
            return self._df._with_dataframe(self._df._df.where(key))
        elif isinstance(key, (list, tuple)):
            # Column selection only
            str_columns = [c for c in key if isinstance(c, str)]
            if str_columns:
                self._df._validate_columns_exist(str_columns, "loc column selection")
            from ...expressions.column import col

            columns = [col(c) if isinstance(c, str) else c for c in key]
            return self._df._with_dataframe(self._df._df.select(*columns))
        else:
            raise TypeError(f"Invalid key type for loc: {type(key)}")


class _AsyncILocIndexer:
    """Indexer for pandas-style iloc accessor (async)."""

    _df: AsyncPandasDataFrame

    def __init__(self, df: AsyncPandasDataFrame):
        self._df = df

    def __getitem__(self, key: Any) -> AsyncPandasDataFrame:
        """Access rows and columns using iloc (integer position).

        Note:
            Full iloc functionality requires materialization.
            Only boolean array filtering is supported for lazy evaluation.
        """
        # For lazy evaluation, we can only support boolean filtering
        if isinstance(key, Column):
            # Boolean condition
            return self._df._with_dataframe(self._df._df.where(key))
        else:
            raise NotImplementedError(
                "iloc positional indexing requires materialization. "
                "Use limit() or boolean filtering instead."
            )



"""Shared helper functions for Polars GroupBy operations.

This module contains shared logic used by both PolarsGroupBy and AsyncPolarsGroupBy
to reduce code duplication and improve maintainability.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, List, Protocol, Tuple, TypeVar, Union

from ...expressions.column import Column, col

if TYPE_CHECKING:
    from ..core.dataframe import DataFrame
    from ..core.async_dataframe import AsyncDataFrame
    from ..groupby.groupby import GroupedDataFrame
    from ..groupby.async_groupby import AsyncGroupedDataFrame
    from ...logical.plan import LogicalPlan
else:
    GroupedDataFrame = Any
    AsyncGroupedDataFrame = Any
    DataFrame = Any
    AsyncDataFrame = Any
    LogicalPlan = Any

# Type variable for generic Polars GroupBy operations
G = TypeVar("G", bound="PolarsGroupByProtocol")

if TYPE_CHECKING:

    class PolarsGroupByProtocol(Protocol):
        """Protocol defining the interface that Polars GroupBy classes must implement."""

        _grouped: Union[GroupedDataFrame, AsyncGroupedDataFrame]

        def agg(
            self, *aggregations: Union[Column, str, dict[str, str]]
        ) -> Union[DataFrame, AsyncDataFrame]:
            """Apply aggregations to the grouped data."""
            ...
else:
    PolarsGroupByProtocol = Any


def build_polars_groupby_agg_with_dict_handling(
    polars_groupby: Any,  # PolarsGroupByProtocol - using Any for mypy compatibility
    *exprs: Union[Column, dict[str, str]],
) -> Tuple[List[Column], Union[GroupedDataFrame, AsyncGroupedDataFrame]]:
    """Build aggregations for Polars GroupBy with dictionary syntax support.

    Args:
        polars_groupby: Polars GroupBy instance
        *exprs: Column expressions or dictionaries mapping column names to function names

    Returns:
        Tuple of (normalized expressions list, grouped dataframe instance)
    """
    # Handle dictionary syntax
    normalized_exprs = []
    for expr in exprs:
        if isinstance(expr, dict):
            # Dictionary syntax: {"column": "function"}
            for col_name, func_name in expr.items():
                agg_col = polars_groupby._grouped._create_aggregation_from_string(
                    col_name, func_name
                )
                normalized_exprs.append(agg_col)
        else:
            # Column expression
            normalized_exprs.append(expr)

    return normalized_exprs, polars_groupby._grouped


def build_polars_groupby_column_aggregation(
    polars_groupby: Any,  # PolarsGroupByProtocol - using Any for mypy compatibility
    func_name: str,
    func_builder: Callable[[Column], Column],
    suffix: str,
    error_msg: str,
) -> List[Column]:
    """Build column-based aggregation for Polars GroupBy methods like mean, sum, etc.

    Args:
        polars_groupby: Polars GroupBy instance
        func_name: Name of the aggregation function (for error messages)
        func_builder: Function that takes a Column and returns an aggregated Column
        suffix: Suffix for aliases (e.g., "_mean", "_sum")
        error_msg: Error message to show if no columns found

    Returns:
        List of aggregation Column expressions

    Raises:
        NotImplementedError: If columns cannot be extracted
        ValueError: If no columns found for aggregation
    """
    # Extract columns from the grouped dataframe
    from ...logical.plan import Aggregate

    grouped = polars_groupby._grouped
    columns: List[str] = []

    # Try to get columns - sync and async have different structures
    if hasattr(grouped, "parent"):
        # Sync GroupBy - has parent DataFrame
        parent_df = grouped.parent
        try:
            columns = parent_df._extract_column_names(parent_df.plan)
        except Exception:
            raise NotImplementedError(
                f"{func_name}() requires accessible columns - use agg() with explicit columns"
            )
    elif hasattr(grouped, "plan") and isinstance(grouped.plan, Aggregate) and grouped.plan.child:
        # Async GroupBy - extract from plan
        from ..core.async_dataframe import AsyncDataFrame

        child_plan = grouped.plan.child
        temp_df = AsyncDataFrame(plan=child_plan, database=getattr(grouped, "database", None))
        try:
            columns = temp_df._extract_column_names(child_plan)
        except Exception:
            raise NotImplementedError(
                f"{func_name}() requires accessible columns - use agg() with explicit columns"
            )
    else:
        raise NotImplementedError(
            f"{func_name}() requires accessible columns - use agg() with explicit columns"
        )

    # Build aggregation list
    agg_list = []
    for col_name in columns:
        try:
            agg_col = func_builder(col(col_name))
            agg_list.append(agg_col.alias(f"{col_name}{suffix}"))
        except Exception:
            pass

    if not agg_list:
        raise ValueError(error_msg)

    return agg_list



"""Custom exception hierarchy."""

from typing import Optional, Sequence


def _levenshtein_distance(s1: str, s2: str) -> int:
    """Calculate Levenshtein distance between two strings."""
    if len(s1) < len(s2):
        return _levenshtein_distance(s2, s1)
    if len(s2) == 0:
        return len(s1)
    previous_row = list(range(len(s2) + 1))
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            insertions = previous_row[j + 1] + 1
            deletions = current_row[j] + 1
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row
    return previous_row[-1]


def _suggest_column_name(column_name: str, available_columns: Sequence[str]) -> str:
    """Suggest similar column names when a column is not found."""
    if not available_columns:
        return f"Column '{column_name}' does not exist. No columns are available in this context."
    # Calculate similarity scores
    scores = [
        (col, _levenshtein_distance(column_name.lower(), col.lower())) for col in available_columns
    ]
    scores.sort(key=lambda x: x[1])
    # Get top 3 suggestions
    suggestions = [col for col, _ in scores[:3] if scores[0][1] <= len(column_name)]
    if suggestions:
        if len(suggestions) == 1:
            return f"Column '{column_name}' does not exist. Did you mean: '{suggestions[0]}'?"
        else:
            suggestions_str = ", ".join(f"'{s}'" for s in suggestions)
            return f"Column '{column_name}' does not exist. Did you mean one of: {suggestions_str}?"
    return (
        f"Column '{column_name}' does not exist. "
        f"Available columns: {', '.join(available_columns[:10])}"
        + ("..." if len(available_columns) > 10 else "")
    )


def _suggest_table_name(table_name: str, available_tables: Sequence[str]) -> str:
    """Suggest similar table names when a table is not found."""
    if not available_tables:
        return f"Table '{table_name}' does not exist. No tables are available in this database."
    # Calculate similarity scores
    scores = [
        (tbl, _levenshtein_distance(table_name.lower(), tbl.lower())) for tbl in available_tables
    ]
    scores.sort(key=lambda x: x[1])
    # Get top 3 suggestions
    suggestions = [tbl for tbl, _ in scores[:3] if scores[0][1] <= len(table_name)]
    if suggestions:
        if len(suggestions) == 1:
            return f"Table '{table_name}' does not exist. Did you mean: '{suggestions[0]}'?"
        else:
            suggestions_str = ", ".join(f"'{s}'" for s in suggestions)
            return f"Table '{table_name}' does not exist. Did you mean one of: {suggestions_str}?"
    return (
        f"Table '{table_name}' does not exist. "
        f"Available tables: {', '.join(available_tables[:10])}"
        + ("..." if len(available_tables) > 10 else "")
    )


class MoltresError(Exception):
    """Base exception for Moltres-specific failures."""

    def __init__(
        self, message: str, suggestion: Optional[str] = None, context: Optional[dict] = None
    ):
        """Initialize exception with message, optional suggestion, and context.

        Args:
            message: Error message
            suggestion: Optional suggestion for fixing the error
            context: Optional dictionary with additional context
        """
        super().__init__(message)
        self.message = message
        self.suggestion = suggestion
        self.context = context or {}

    def __str__(self) -> str:
        """Return formatted error message with suggestion if available."""
        msg = self.message
        if self.suggestion:
            msg += f"\n\nSuggestion: {self.suggestion}"
        if self.context:
            context_str = ", ".join(f"{k}={v}" for k, v in self.context.items())
            msg += f"\n\nContext: {context_str}"
        return msg


class CompilationError(MoltresError):
    """Raised when a logical plan cannot be converted into SQL."""

    def __init__(
        self, message: str, suggestion: Optional[str] = None, context: Optional[dict] = None
    ):
        """Initialize compilation error.

        Common suggestions:
        - Check if the operation is supported for your SQL dialect
        - Verify column names and table references
        - Ensure join conditions are properly specified
        """
        # Auto-generate suggestions for common errors
        if suggestion is None:
            if "not supported" in message.lower() or "unsupported" in message.lower():
                suggestion = (
                    "This operation may not be supported for your SQL dialect. "
                    "Check the documentation for dialect-specific features."
                )
            elif "join" in message.lower() and "condition" in message.lower():
                suggestion = (
                    "Joins require either an 'on' parameter with column pairs "
                    "or a 'condition' parameter with a Column expression. "
                    "Example: df.join(other, on=[col('left_col') == col('right_col')])"
                )
            elif "subquery" in message.lower():
                suggestion = (
                    "Subqueries require a DataFrame with a logical plan. "
                    "Make sure you're passing a DataFrame, not a list or other type."
                )
        super().__init__(message, suggestion, context)


class ExecutionError(MoltresError):
    """Raised when SQL execution fails."""

    def __init__(
        self, message: str, suggestion: Optional[str] = None, context: Optional[dict] = None
    ):
        """Initialize execution error.

        Common suggestions:
        - Check SQL syntax and table/column names
        - Verify database connection is active
        - Check for data type mismatches
        """
        if suggestion is None:
            if "no such table" in message.lower() or "relation" in message.lower():
                suggestion = (
                    "The table does not exist. Check the table name spelling and "
                    "ensure the table has been created. Use db.table('name') to verify."
                )
                # Try to extract table name and suggest similar ones
                context = context or {}
                if "table_name" in context and "available_tables" in context:
                    suggestion = _suggest_table_name(
                        context["table_name"], context["available_tables"]
                    )
            elif "no such column" in message.lower():
                suggestion = (
                    "The column does not exist. Check the column name spelling. "
                    "Use df.select() to see available columns."
                )
                # Try to extract column name and suggest similar ones
                context = context or {}
                if "column_name" in context and "available_columns" in context:
                    suggestion = _suggest_column_name(
                        context["column_name"], context["available_columns"]
                    )
            elif "syntax error" in message.lower():
                suggestion = (
                    "There's a SQL syntax error. Check your query structure. "
                    "Use df.to_sql() to see the generated SQL."
                )
        super().__init__(message, suggestion, context)


class ValidationError(MoltresError):
    """Raised when input validation fails."""

    def __init__(
        self, message: str, suggestion: Optional[str] = None, context: Optional[dict] = None
    ):
        """Initialize validation error.

        Common suggestions:
        - Check input types and formats
        - Verify required parameters are provided
        - Check value ranges and constraints
        """
        if suggestion is None:
            if "column name" in message.lower():
                suggestion = (
                    "Column names must be valid identifiers (letters, digits, underscores). "
                    "Avoid SQL keywords and special characters."
                )
            elif "required" in message.lower() or "missing" in message.lower():
                suggestion = "Check that all required parameters are provided."
        super().__init__(message, suggestion, context)


class SchemaError(MoltresError):
    """Raised when schema-related operations fail."""


class DatabaseConnectionError(MoltresError):
    """Raised when database connection operations fail."""

    def __init__(
        self, message: str, suggestion: Optional[str] = None, context: Optional[dict] = None
    ):
        """Initialize connection error."""
        if suggestion is None:
            suggestion = (
                "Check your database connection string and ensure the database server is running. "
                "Verify network connectivity and credentials."
            )
        super().__init__(message, suggestion, context)


class QueryTimeoutError(ExecutionError):
    """Raised when a query exceeds the configured timeout."""

    def __init__(
        self, message: str, timeout: Optional[float] = None, context: Optional[dict] = None
    ):
        """Initialize query timeout error."""
        suggestion = (
            "The query exceeded the timeout limit. Consider:\n"
            "  - Optimizing the query (add indexes, reduce data scanned)\n"
            "  - Increasing the timeout via query_timeout configuration\n"
            "  - Breaking the query into smaller chunks"
        )
        if timeout is not None:
            context = context or {}
            context["timeout_seconds"] = timeout
        super().__init__(message, suggestion, context)


class ConnectionPoolError(DatabaseConnectionError):
    """Raised when connection pool operations fail."""

    def __init__(
        self, message: str, suggestion: Optional[str] = None, context: Optional[dict] = None
    ):
        """Initialize connection pool error."""
        if suggestion is None:
            suggestion = (
                "Connection pool error. Consider:\n"
                "  - Increasing pool_size and max_overflow\n"
                "  - Checking for connection leaks (unclosed connections)\n"
                "  - Verifying database server capacity"
            )
        super().__init__(message, suggestion, context)


class TransactionError(ExecutionError):
    """Raised when transaction operations fail."""

    def __init__(
        self, message: str, suggestion: Optional[str] = None, context: Optional[dict] = None
    ):
        """Initialize transaction error."""
        if suggestion is None:
            suggestion = (
                "Transaction error. Ensure:\n"
                "  - All operations in the transaction are valid\n"
                "  - No conflicting locks exist\n"
                "  - Database supports the transaction isolation level"
            )
        super().__init__(message, suggestion, context)


class UnsupportedOperationError(MoltresError):
    """Raised when an unsupported operation is attempted."""

    def __init__(
        self, message: str, suggestion: Optional[str] = None, context: Optional[dict] = None
    ):
        """Initialize unsupported operation error."""
        if suggestion is None:
            suggestion = (
                "This operation is not supported. Check the documentation "
                "for supported operations and alternatives."
            )
        super().__init__(message, suggestion, context)


class PandasAPIError(MoltresError):
    """Raised when pandas-style API operations fail."""

    def __init__(
        self, message: str, suggestion: Optional[str] = None, context: Optional[dict] = None
    ):
        """Initialize pandas API error."""
        if suggestion is None:
            suggestion = (
                "This pandas-style operation encountered an error. "
                "Check that column names are correct and operations are valid."
            )
        super().__init__(message, suggestion, context)


class ColumnNotFoundError(ValidationError):
    """Raised when a column is not found in a :class:`DataFrame`."""

    def __init__(
        self,
        column_name: str,
        available_columns: Sequence[str],
        suggestion: Optional[str] = None,
        context: Optional[dict] = None,
    ):
        """Initialize column not found error."""
        message = f"Column '{column_name}' not found"
        if suggestion is None:
            suggestion = _suggest_column_name(column_name, available_columns)
        context = context or {}
        context["column_name"] = column_name
        context["available_columns"] = available_columns
        super().__init__(message, suggestion, context)



"""Compile expression trees into SQLAlchemy column expressions."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Optional, Union
from typing import cast as typing_cast

from sqlalchemy import (
    func,
    case as sa_case,
    null,
    literal,
    select,
)
from sqlalchemy.sql import Select, ColumnElement

from ..engine.dialects import DialectSpec
from ..expressions.column import Column
from ..utils.exceptions import CompilationError, ValidationError
from .builders import quote_identifier

if TYPE_CHECKING:
    from sqlalchemy.sql import Subquery, TableClause
    from .plan_compiler import SQLCompiler

logger = logging.getLogger(__name__)


class ExpressionCompiler:
    """Compile expression trees into SQLAlchemy column expressions."""

    def __init__(self, dialect: DialectSpec, plan_compiler: Optional[SQLCompiler] = None):
        self.dialect = dialect
        self._table_cache: dict[str, Any] = {}
        self._current_subq: Subquery | None = None
        self._join_info: Optional[dict[str, str]] = None
        self._plan_compiler = plan_compiler

    def compile_expr(self, expression: Column) -> ColumnElement:
        """Compile a :class:`Column` expression to a SQLAlchemy column expression."""
        return self._compile(expression)

    def emit(self, expression: Column) -> str:
        """Compile a :class:`Column` expression to a SQL string.

        Args:
            expression: :class:`Column` expression to compile

        Returns:
            SQL string representation of the expression
        """
        compiled = self.compile_expr(expression)
        # Convert SQLAlchemy column element to SQL string
        return str(compiled.compile(compile_kwargs={"literal_binds": True}))

    def _compile(self, expression: Union[Column, str, Any]) -> ColumnElement:
        """Compile a :class:`Column` expression to a SQLAlchemy column expression."""
        # Import here to ensure it's available throughout the method
        from sqlalchemy import column as sa_column, literal_column
        from sqlalchemy.sql import ColumnElement

        # Handle string literals directly
        if isinstance(expression, str):
            return literal(expression)

        # Must be a Column expression at this point
        if not isinstance(expression, Column):
            raise CompilationError(f"Expected Column expression, got {type(expression)}")

        op = expression.op

        if op == "star":
            # "*" means select all columns - use literal_column("*")
            star_result: ColumnElement[Any] = literal_column("*")
            if expression._alias:
                star_result = star_result.label(expression._alias)
            return star_result

        if op == "column":
            col_name = expression.args[0]
            # Type narrowing: column op always has a string as first arg
            if not isinstance(col_name, str):
                raise TypeError(f"Expected string for column name, got {type(col_name).__name__}")
            # Validate column name to prevent SQL injection
            try:
                # This will raise ValidationError if column name is invalid
                quote_identifier(col_name, self.dialect.quote_char)
            except ValidationError:
                # Re-raise with clearer context
                raise ValidationError(
                    f"Invalid column name: {col_name!r}. "
                    "Column names may only contain letters, digits, underscores, and dots."
                ) from None

            # Handle qualified column names (table.column)

            sa_col: ColumnElement[Any]
            if "." in col_name:
                parts = col_name.split(".", 1)
                table_name = parts[0]
                col_name = parts[1]  # Extract column name
                # If we have a current subquery (from join), try to access column from it
                # After a join with SELECT *, columns lose table qualification
                if self._current_subq is not None:
                    try:
                        # Try to access column from subquery's column collection
                        # The column name in the subquery is just the column name, not table.column
                        sa_col = typing_cast(ColumnElement[Any], self._current_subq.c[col_name])
                    except (KeyError, AttributeError, TypeError):
                        # Column not found in subquery, try using table-qualified literal
                        # This might work if the subquery preserves table info
                        quote = self.dialect.quote_char
                        sa_col = typing_cast(
                            ColumnElement[Any],
                            literal_column(f"{quote}{table_name}{quote}.{quote}{col_name}{quote}"),
                        )
                else:
                    # No subquery context, use qualified literal column
                    quote = self.dialect.quote_char
                    sa_col = typing_cast(
                        ColumnElement[Any],
                        literal_column(f"{quote}{table_name}{quote}.{quote}{col_name}{quote}"),
                    )
            else:
                # Unqualified column - will be resolved in context
                sa_col = typing_cast(ColumnElement[Any], sa_column(col_name))
            if expression._alias:
                # label() returns a Label which is a ColumnElement, but mypy sees it as Any
                sa_col = typing_cast(ColumnElement[Any], sa_col.label(expression._alias))
            return sa_col

        if op == "literal":
            value = expression.args[0]
            sa_lit: ColumnElement[Any] = literal(value)
            if expression._alias:
                sa_lit = sa_lit.label(expression._alias)
            return sa_lit

        # Try math compiler for basic arithmetic and math functions
        if op in (
            "add",
            "sub",
            "mul",
            "div",
            "floor_div",
            "round",
            "floor",
            "ceil",
            "abs",
            "sqrt",
            "exp",
            "log",
            "log10",
            "log2",
            "sin",
            "cos",
            "tan",
            "asin",
            "acos",
            "atan",
            "atan2",
            "signum",
            "sign",
            "hypot",
        ):
            from .expression_compilers.math import compile_math_operation

            result = compile_math_operation(self, op, expression)
            if result is not None:
                return result
        if op == "eq":
            left, right = expression.args
            result = self._compile(left) == self._compile(right)
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "ne":
            left, right = expression.args
            result = self._compile(left) != self._compile(right)
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "lt":
            left, right = expression.args
            result = self._compile(left) < self._compile(right)
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "le":
            left, right = expression.args
            result = self._compile(left) <= self._compile(right)
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "gt":
            left, right = expression.args
            result = self._compile(left) > self._compile(right)
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "ge":
            left, right = expression.args
            result = self._compile(left) >= self._compile(right)
            if expression._alias:
                result = result.label(expression._alias)
            return result

        # Date/time functions
        if op == "year":
            result = func.extract("year", self._compile(expression.args[0]))
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "month":
            result = func.extract("month", self._compile(expression.args[0]))
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "day":
            result = func.extract("day", self._compile(expression.args[0]))
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "dayofweek":
            result = func.extract("dow", self._compile(expression.args[0]))  # Day of week
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "hour":
            result = func.extract("hour", self._compile(expression.args[0]))
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "minute":
            result = func.extract("minute", self._compile(expression.args[0]))
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "second":
            result = func.extract("second", self._compile(expression.args[0]))
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "date_format":
            col_expr = self._compile(expression.args[0])
            format_str = expression.args[1]
            # Use to_char for PostgreSQL, DATE_FORMAT for MySQL, strftime for SQLite
            result = func.to_char(col_expr, format_str)
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "to_date":
            col_expr = self._compile(expression.args[0])
            if len(expression.args) > 1:
                format_str = expression.args[1]
                result = func.to_date(col_expr, format_str)
            else:
                result = func.to_date(col_expr)
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "current_date":
            result = func.current_date()
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "current_timestamp":
            result = func.now()
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "datediff":
            end = self._compile(expression.args[0])
            start = self._compile(expression.args[1])
            result = end - start  # Simplified - actual datediff varies by dialect
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "date_add":
            col_expr = self._compile(expression.args[0])
            interval_str = expression.args[1]  # e.g., "1 DAY", "2 MONTH"
            if not isinstance(interval_str, str):
                raise TypeError(f"Expected string for interval, got {type(interval_str).__name__}")
            from sqlalchemy import literal_column

            # Parse interval string (format: "N UNIT" where N is number and UNIT is DAY, MONTH, YEAR, HOUR, etc.)
            parts = interval_str.split()
            if len(parts) != 2:
                raise CompilationError(
                    f"Invalid interval format: {interval_str}. Expected format: 'N UNIT' (e.g., '1 DAY')"
                )
            num, unit = parts
            unit_upper = unit.upper()

            # For PostgreSQL/DuckDB, use INTERVAL literal
            if self.dialect.name in ("postgresql", "duckdb"):
                interval_col: ColumnElement[Any] = literal_column(f"INTERVAL '{interval_str}'")
                result = col_expr + interval_col
            elif self.dialect.name == "mysql":
                # MySQL uses DATE_ADD with INTERVAL
                result = func.date_add(col_expr, literal_column(f"INTERVAL {num} {unit_upper}"))
            else:
                # SQLite: use datetime() function with modifier
                # SQLite format: datetime(col, '+1 day'), datetime(col, '-1 hour')
                modifier = f"+{num} {unit_upper.lower()}"
                result = func.datetime(col_expr, modifier)
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "date_sub":
            col_expr = self._compile(expression.args[0])
            interval_str = expression.args[1]  # e.g., "1 DAY", "2 MONTH"
            if not isinstance(interval_str, str):
                raise TypeError(f"Expected string for interval, got {type(interval_str).__name__}")
            from sqlalchemy import literal_column

            # Parse interval string
            parts = interval_str.split()
            if len(parts) != 2:
                raise CompilationError(
                    f"Invalid interval format: {interval_str}. Expected format: 'N UNIT' (e.g., '1 DAY')"
                )
            num, unit = parts
            unit_upper = unit.upper()

            # For PostgreSQL/DuckDB, use INTERVAL literal
            if self.dialect.name in ("postgresql", "duckdb"):
                interval_expr: ColumnElement[Any] = literal_column(f"INTERVAL '{interval_str}'")
                result = col_expr - interval_expr
            elif self.dialect.name == "mysql":
                # MySQL uses DATE_SUB with INTERVAL
                result = func.date_sub(col_expr, literal_column(f"INTERVAL {num} {unit_upper}"))
            else:
                # SQLite: use datetime() function with modifier
                modifier = f"-{num} {unit_upper.lower()}"
                result = func.datetime(col_expr, modifier)
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "add_months":
            col_expr = self._compile(expression.args[0])
            num_months = expression.args[1]
            if not isinstance(num_months, (int, float)):
                raise TypeError(f"Expected number for months, got {type(num_months).__name__}")
            # Use SQLAlchemy's interval handling
            # Try to use make_interval if available (PostgreSQL), otherwise use date_add
            try:
                interval_months: ColumnElement[Any] = func.make_interval(
                    months=abs(int(num_months))
                )
                if num_months >= 0:
                    result = col_expr + interval_months
                else:
                    result = col_expr - interval_months
            except (NotImplementedError, AttributeError, TypeError) as e:
                # Fallback: use date_add function (MySQL/SQLite compatible)
                # This is a simplified fallback
                logger.debug(
                    "make_interval not available for dialect, using date_add fallback: %s", e
                )
                result = func.date_add(col_expr, literal(num_months))
            except Exception as e:
                # Catch any other unexpected errors and log them before falling back
                logger.warning(
                    "Unexpected error using make_interval, falling back to date_add: %s", e
                )
                result = func.date_add(col_expr, literal(num_months))
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "to_timestamp":
            col_expr = self._compile(expression.args[0])
            if len(expression.args) > 1:
                format_str = expression.args[1]
                if not isinstance(format_str, str):
                    raise TypeError(f"Expected string for format, got {type(format_str).__name__}")
                if self.dialect.name == "duckdb":
                    # DuckDB uses strptime for parsing with format (uses %Y format, not yyyy)
                    from sqlalchemy import literal_column

                    # Convert PySpark format to strptime format
                    duckdb_format = (
                        format_str.replace("yyyy", "%Y")
                        .replace("MM", "%m")
                        .replace("dd", "%d")
                        .replace("HH", "%H")
                        .replace("mm", "%M")
                        .replace("ss", "%S")
                    )
                    result = literal_column(f"strptime({col_expr}, '{duckdb_format}')")
                else:
                    result = func.to_timestamp(col_expr, format_str)
            else:
                if self.dialect.name == "duckdb":
                    # DuckDB's to_timestamp expects a numeric unix timestamp
                    result = func.to_timestamp(col_expr)
                else:
                    result = func.to_timestamp(col_expr)
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "unix_timestamp":
            from sqlalchemy import literal_column

            if len(expression.args) == 0:
                # Current Unix timestamp
                if self.dialect.name == "postgresql":
                    result = func.extract("epoch", func.now())
                elif self.dialect.name == "mysql":
                    result = func.unix_timestamp()
                elif self.dialect.name == "duckdb":
                    # DuckDB: extract(epoch from now())
                    result = func.extract("epoch", func.now())
                else:
                    # SQLite: strftime('%s', 'now')
                    result = func.strftime("%s", "now")
            elif len(expression.args) == 1:
                col_expr = self._compile(expression.args[0])
                if self.dialect.name == "postgresql":
                    result = func.extract("epoch", col_expr)
                elif self.dialect.name == "mysql":
                    result = func.unix_timestamp(col_expr)
                elif self.dialect.name == "duckdb":
                    # DuckDB: extract(epoch from col)
                    result = func.extract("epoch", col_expr)
                else:
                    # SQLite: strftime('%s', col)
                    result = func.strftime("%s", col_expr)
            else:
                col_expr = self._compile(expression.args[0])
                format_str = expression.args[1]
                if self.dialect.name == "postgresql":
                    # Parse with format then extract epoch
                    parsed = func.to_timestamp(col_expr, format_str)
                    result = func.extract("epoch", parsed)
                elif self.dialect.name == "mysql":
                    result = func.unix_timestamp(col_expr, format_str)
                elif self.dialect.name == "duckdb":
                    # DuckDB: parse with format then extract epoch
                    parsed = func.to_timestamp(col_expr, format_str)
                    result = func.extract("epoch", parsed)
                else:
                    # SQLite: requires parsing first
                    result = func.strftime("%s", func.datetime(col_expr, format_str))
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "from_unixtime":
            col_expr = self._compile(expression.args[0])
            if len(expression.args) > 1:
                format_str = expression.args[1]
                if not isinstance(format_str, str):
                    raise TypeError(f"Expected string for format, got {type(format_str).__name__}")
                if self.dialect.name == "postgresql":
                    result = func.to_char(func.to_timestamp(col_expr), format_str)
                elif self.dialect.name == "mysql":
                    result = func.from_unixtime(col_expr, format_str)
                elif self.dialect.name == "duckdb":
                    # DuckDB: to_timestamp(unix_time) then format with strftime
                    # DuckDB's strftime is strftime(timestamp, format)
                    from sqlalchemy import literal_column

                    duckdb_format = (
                        format_str.replace("yyyy", "%Y")
                        .replace("MM", "%m")
                        .replace("dd", "%d")
                        .replace("HH", "%H")
                        .replace("mm", "%M")
                        .replace("ss", "%S")
                    )
                    result = literal_column(
                        f"strftime(to_timestamp({col_expr}), '{duckdb_format}')"
                    )
                else:
                    # SQLite: datetime(unix_time, 'unixepoch')
                    result = func.strftime(format_str, func.datetime(col_expr, "unixepoch"))
            else:
                if self.dialect.name in ("postgresql", "duckdb"):
                    result = func.to_timestamp(col_expr)
                elif self.dialect.name == "mysql":
                    result = func.from_unixtime(col_expr)
                else:
                    # SQLite: datetime(unix_time, 'unixepoch')
                    result = func.datetime(col_expr, "unixepoch")
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "date_trunc":
            trunc_unit_arg = expression.args[0]
            if not isinstance(trunc_unit_arg, str):
                raise TypeError(
                    f"Expected string for date_trunc unit, got {type(trunc_unit_arg).__name__}"
                )
            trunc_unit: str = trunc_unit_arg
            col_expr = self._compile(expression.args[1])
            if self.dialect.name == "postgresql":
                result = func.date_trunc(trunc_unit, col_expr)
            elif self.dialect.name == "mysql":
                # MySQL: use DATE_FORMAT with truncation
                from sqlalchemy import literal_column

                unit_map = {
                    "year": "%Y-01-01",
                    "month": "%Y-%m-01",
                    "day": "%Y-%m-%d",
                    "hour": "%Y-%m-%d %H:00:00",
                    "minute": "%Y-%m-%d %H:%i:00",
                    "second": "%Y-%m-%d %H:%i:%s",
                }
                format_str = unit_map.get(unit.lower(), "%Y-%m-%d")
                result = func.date_format(col_expr, format_str)
            else:
                # SQLite: use strftime
                unit_map = {
                    "year": "%Y-01-01",
                    "month": "%Y-%m-01",
                    "day": "%Y-%m-%d",
                    "hour": "%Y-%m-%d %H:00:00",
                    "minute": "%Y-%m-%d %H:%M:00",
                    "second": "%Y-%m-%d %H:%M:%S",
                }
                format_str = unit_map.get(unit.lower(), "%Y-%m-%d")
                result = func.strftime(format_str, col_expr)
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "quarter":
            col_expr = self._compile(expression.args[0])
            if self.dialect.name == "postgresql":
                result = func.extract("quarter", col_expr)
            elif self.dialect.name == "mysql":
                result = func.quarter(col_expr)
            else:
                # SQLite: strftime('%m') then calculate quarter
                from sqlalchemy import literal_column, types as sa_types

                month = func.cast(func.strftime("%m", col_expr), sa_types.Integer)
                result = sa_case(
                    (month <= 3, literal(1)),
                    (month <= 6, literal(2)),
                    (month <= 9, literal(3)),
                    else_=literal(4),
                )
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "weekofyear" or op == "week":
            col_expr = self._compile(expression.args[0])
            if self.dialect.name == "postgresql":
                result = func.extract("week", col_expr)
            elif self.dialect.name == "mysql":
                result = func.week(col_expr)
            else:
                # SQLite: strftime('%W')
                from sqlalchemy import types as sa_types

                result = func.cast(func.strftime("%W", col_expr), sa_types.Integer) + 1
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "dayofyear":
            col_expr = self._compile(expression.args[0])
            if self.dialect.name == "postgresql":
                result = func.extract("doy", col_expr)
            elif self.dialect.name == "mysql":
                result = func.dayofyear(col_expr)
            else:
                # SQLite: strftime('%j')
                from sqlalchemy import types as sa_types

                result = func.cast(func.strftime("%j", col_expr), sa_types.Integer)
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "last_day":
            col_expr = self._compile(expression.args[0])
            if self.dialect.name == "postgresql":
                # PostgreSQL: date_trunc('month', date) + interval '1 month' - interval '1 day'
                from sqlalchemy import literal_column

                result = (
                    func.date_trunc("month", col_expr)
                    + literal_column("INTERVAL '1 month'")
                    - literal_column("INTERVAL '1 day'")
                )
            elif self.dialect.name in ("mysql", "duckdb"):
                result = func.last_day(col_expr)
            else:
                # SQLite: requires workaround
                from sqlalchemy import literal_column

                # Use date() with month+1, day=0 trick
                result = func.date(col_expr, "+1 month", "-1 day")
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "months_between":
            date1 = self._compile(expression.args[0])
            date2 = self._compile(expression.args[1])
            if self.dialect.name == "postgresql":
                # PostgreSQL: EXTRACT(EPOCH FROM (date1 - date2)) / (30.0 * 86400)
                from sqlalchemy import literal_column

                result = func.extract("epoch", date1 - date2) / literal(30.0 * 86400)
            elif self.dialect.name == "mysql":
                result = func.timestampdiff("month", date2, date1)
            elif self.dialect.name == "duckdb":
                # DuckDB: Calculate months using date difference
                from sqlalchemy import literal_column

                # Use datediff with month unit
                result = func.datediff("month", date2, date1)
            else:
                # SQLite: requires calculation
                from sqlalchemy import literal_column

                # Approximate: (julianday(date1) - julianday(date2)) / 30.0
                result = (func.julianday(date1) - func.julianday(date2)) / literal(30.0)
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "mod":
            left, right = expression.args
            return func.mod(self._compile(left), self._compile(right))
        if op == "pow":
            base, exp = expression.args[:2]
            result = func.power(self._compile(base), self._compile(exp))
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "neg":
            return -self._compile(expression.args[0])
        # Try logical compiler for logical operations
        if op in ("and", "or", "not", "between", "in", "case_when"):
            from .expression_compilers.logical import compile_logical_operation

            result = compile_logical_operation(self, op, expression)
            if result is not None:
                return result
        if op == "in_subquery":
            # IN with subquery: col("id").isin(df.select("id"))
            value, subquery_plan = expression.args
            from ..logical.plan import LogicalPlan

            if not isinstance(subquery_plan, LogicalPlan):
                raise TypeError(
                    f"Expected LogicalPlan for subquery, got {type(subquery_plan).__name__}"
                )
            if self._plan_compiler is None:
                raise CompilationError("Subquery compilation requires plan compiler")
            subquery_stmt = self._plan_compiler._compile_plan(subquery_plan)
            return self._compile(value).in_(subquery_stmt)
        if op == "scalar_subquery":
            # Scalar subquery in SELECT: scalar_subquery(df.select())
            (subquery_plan,) = expression.args
            from ..logical.plan import LogicalPlan

            if not isinstance(subquery_plan, LogicalPlan):
                raise TypeError(
                    f"Expected LogicalPlan for subquery, got {type(subquery_plan).__name__}"
                )
            if self._plan_compiler is None:
                raise CompilationError("Subquery compilation requires plan compiler")
            subquery_stmt = self._plan_compiler._compile_plan(subquery_plan)
            # SQLAlchemy scalar_subquery() - must return single row/column
            # Use select().scalar_subquery() method
            from sqlalchemy.sql import ColumnElement as ColumnElementType

            if isinstance(subquery_stmt, Select):
                scalar_result: ColumnElementType[Any] = subquery_stmt.scalar_subquery()
            else:
                # Fallback: wrap in select
                scalar_result = select(subquery_stmt).scalar_subquery()
            if expression._alias:
                scalar_result = scalar_result.label(expression._alias)
            return scalar_result
        if op == "exists":
            # EXISTS subquery: exists(df.select())
            (subquery_plan,) = expression.args
            from ..logical.plan import LogicalPlan

            if not isinstance(subquery_plan, LogicalPlan):
                raise TypeError(
                    f"Expected LogicalPlan for subquery, got {type(subquery_plan).__name__}"
                )
            if self._plan_compiler is None:
                raise CompilationError("Subquery compilation requires plan compiler")
            subquery_stmt = self._plan_compiler._compile_plan(subquery_plan)
            # SQLAlchemy's exists() function
            from sqlalchemy import exists as sa_exists

            result = sa_exists(subquery_stmt)
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "not_exists":
            # NOT EXISTS subquery: not_exists(df.select())
            (subquery_plan,) = expression.args
            from ..logical.plan import LogicalPlan

            if not isinstance(subquery_plan, LogicalPlan):
                raise TypeError(
                    f"Expected LogicalPlan for subquery, got {type(subquery_plan).__name__}"
                )
            if self._plan_compiler is None:
                raise CompilationError("Subquery compilation requires plan compiler")
            subquery_stmt = self._plan_compiler._compile_plan(subquery_plan)
            # SQLAlchemy's exists() function with negation
            from sqlalchemy import exists as sa_exists

            result = ~sa_exists(subquery_stmt)
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "like":
            left, pattern = expression.args
            # Pattern is stored as a string - SQLAlchemy's like() accepts string directly
            compiled_left = self._compile(left)
            result = compiled_left.like(pattern)  # pattern is a string
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "ilike":
            left, pattern = expression.args
            # Pattern is stored as a string - SQLAlchemy's ilike() accepts string directly
            compiled_left = self._compile(left)
            result = compiled_left.ilike(pattern)  # pattern is a string
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "contains":
            column, substring = expression.args
            # substring might be a Column or a string
            if isinstance(substring, Column):
                pattern = func.concat(literal("%"), self._compile(substring), literal("%"))
            else:
                pattern = f"%{substring}%"
            return self._compile(column).like(pattern)
        if op == "startswith":
            column, prefix = expression.args
            if isinstance(prefix, Column):
                pattern = func.concat(self._compile(prefix), literal("%"))
            else:
                pattern = f"{prefix}%"
            return self._compile(column).like(pattern)
        if op == "endswith":
            column, suffix = expression.args
            if isinstance(suffix, Column):
                pattern = func.concat(literal("%"), self._compile(suffix))
            else:
                pattern = f"%{suffix}"
            return self._compile(column).like(pattern)
        # Try type casting compiler
        if op == "cast":
            from .expression_compilers.type_casting import compile_type_casting_operation

            result = compile_type_casting_operation(self, op, expression)
            if result is not None:
                return result
        if op == "is_null":
            return self._compile(expression.args[0]).is_(null())
        if op == "is_not_null":
            return self._compile(expression.args[0]).isnot(null())
        if op == "isnan":
            # NaN check - SQL doesn't have direct isnan, use IS NULL or comparison
            # This is a simplified implementation
            col_expr = self._compile(expression.args[0])
            result = col_expr.is_(null())  # Simplified - actual NaN check varies by dialect
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "isinf":
            # Infinity check - SQL doesn't have direct isinf
            # This is a simplified implementation
            col_expr = self._compile(expression.args[0])
            # Use a comparison that would never be true for finite numbers
            # This is dialect-specific and simplified
            result = (col_expr == literal(float("inf"))) | (col_expr == literal(float("-inf")))
            if expression._alias:
                result = result.label(expression._alias)
            return result
        # Try category-specific compilers
        if op.startswith("agg_"):
            from .expression_compilers.aggregation import compile_aggregation

            result = compile_aggregation(self, op, expression)
            if result is not None:
                return result

        if op in (
            "coalesce",
            "concat",
            "upper",
            "lower",
            "substring",
            "trim",
            "ltrim",
            "rtrim",
            "initcap",
            "instr",
            "locate",
            "translate",
            "regexp_extract",
            "regexp_replace",
            "split",
            "replace",
            "length",
            "lpad",
            "rpad",
            "greatest",
            "least",
        ):
            from .expression_compilers.string import compile_string_operation

            result = compile_string_operation(self, op, expression)
            if result is not None:
                return result

        if op in (
            "year",
            "month",
            "day",
            "dayofweek",
            "hour",
            "minute",
            "second",
            "date_format",
            "to_date",
            "current_date",
            "current_timestamp",
            "datediff",
            "date_add",
            "date_sub",
            "add_months",
            "to_timestamp",
            "unix_timestamp",
            "from_unixtime",
            "date_trunc",
            "quarter",
            "weekofyear",
            "week",
            "dayofyear",
            "last_day",
            "months_between",
        ):
            from .expression_compilers.datetime import compile_datetime_operation

            result = compile_datetime_operation(self, op, expression)
            if result is not None:
                return result

        # Try category-specific compilers
        if op.startswith("agg_"):
            from .expression_compilers.aggregation import compile_aggregation

            result = compile_aggregation(self, op, expression)
            if result is not None:
                return result

        if op in (
            "coalesce",
            "concat",
            "upper",
            "lower",
            "substring",
            "trim",
            "ltrim",
            "rtrim",
            "initcap",
            "instr",
            "locate",
            "translate",
            "regexp_extract",
            "regexp_replace",
            "split",
            "replace",
            "length",
            "lpad",
            "rpad",
            "greatest",
            "least",
        ):
            from .expression_compilers.string import compile_string_operation

            result = compile_string_operation(self, op, expression)
            if result is not None:
                return result

        if op in (
            "year",
            "month",
            "day",
            "dayofweek",
            "hour",
            "minute",
            "second",
            "date_format",
            "to_date",
            "current_date",
            "current_timestamp",
            "datediff",
            "date_add",
            "date_sub",
            "add_months",
            "to_timestamp",
            "unix_timestamp",
            "from_unixtime",
            "date_trunc",
            "quarter",
            "weekofyear",
            "week",
            "dayofyear",
            "last_day",
            "months_between",
        ):
            from .expression_compilers.datetime import compile_datetime_operation

            result = compile_datetime_operation(self, op, expression)
            if result is not None:
                return result

        # Try JSON compiler
        if op in ("json_extract", "json_tuple", "json_array_length", "from_json", "to_json"):
            from .expression_compilers.json import compile_json_operation

            result = compile_json_operation(self, op, expression)
            if result is not None:
                return result

        # Array functions
        if op == "array":
            args = [self._compile(arg) for arg in expression.args]
            # Use dialect-specific array construction
            # PostgreSQL: ARRAY[...]
            # SQLite: json_array(...) or string_to_array
            # MySQL: JSON_ARRAY(...)
            if self.dialect.name == "postgresql":
                # PostgreSQL uses ARRAY[...] syntax
                # Use literal_column to generate ARRAY[arg1, arg2, ...] directly
                from sqlalchemy import literal_column

                # Build ARRAY literal by compiling each argument
                array_elements = []
                for arg in args:
                    if hasattr(arg, "compile"):
                        # Compile with literal_binds to get the actual value
                        compiled = arg.compile(compile_kwargs={"literal_binds": True})
                        array_elements.append(str(compiled))
                    else:
                        array_elements.append(str(arg))
                result = literal_column(f"ARRAY[{', '.join(array_elements)}]")
            elif self.dialect.name == "sqlite":
                # SQLite doesn't have native arrays, use JSON array
                result = func.json_array(*args)
            elif self.dialect.name == "mysql":
                # MySQL: Use JSON_ARRAY() with literal values
                # MySQL's JSON_ARRAY doesn't work well with bound parameters
                # Build JSON_ARRAY literal by compiling arguments with literal_binds
                from sqlalchemy import literal_column

                array_elements = []
                for arg in args:
                    if hasattr(arg, "compile"):
                        # Compile with literal_binds to get the actual value
                        compiled = arg.compile(compile_kwargs={"literal_binds": True})
                        array_elements.append(str(compiled))
                    else:
                        array_elements.append(str(arg))
                result = literal_column(f"JSON_ARRAY({', '.join(array_elements)})")
            else:
                # Generic fallback
                result = func.json_array(*args)
            if expression._alias:
                result = result.label(expression._alias)
            return result

        if op == "array_length":
            col_expr = self._compile(expression.args[0])
            # Use dialect-specific array length
            # PostgreSQL: array_length(array, 1)
            # DuckDB: array_length(array) - no dimension argument
            # SQLite: json_array_length(json_array)
            # MySQL: JSON_LENGTH(json_array)
            if self.dialect.name == "postgresql":
                result = func.array_length(col_expr, 1)
            elif self.dialect.name == "duckdb":
                result = func.array_length(col_expr)
            elif self.dialect.name == "sqlite":
                result = func.json_array_length(col_expr)
            else:
                result = func.json_length(col_expr)
            if expression._alias:
                result = result.label(expression._alias)
            return result

        if op == "array_contains":
            col_expr = self._compile(expression.args[0])
            value_expr = self._compile(expression.args[1])
            # Use dialect-specific array contains
            # PostgreSQL: value = ANY(array)
            # SQLite: JSON_CONTAINS (if available) or json_each
            # MySQL: JSON_CONTAINS(json_array, value)
            if self.dialect.name in ("postgresql", "duckdb"):
                # PostgreSQL/DuckDB uses = ANY(array)
                from sqlalchemy import any_

                result = value_expr == any_(col_expr)
            elif self.dialect.name == "mysql":
                # MySQL: JSON_CONTAINS(json_doc, val[, path])
                # The value needs to be a JSON value
                # Use CAST(value AS JSON) to convert the value to JSON type
                from sqlalchemy import cast
                from sqlalchemy.dialects.mysql import JSON as MySQLJSON

                json_value = cast(value_expr, MySQLJSON)
                result = func.json_contains(col_expr, json_value)
            else:
                # SQLite and others - use JSON_CONTAINS if available
                # For SQLite, we'll use a workaround with json_array_length
                try:
                    result = func.json_contains(col_expr, value_expr)
                except (NotImplementedError, AttributeError, TypeError) as e:
                    # Fallback for SQLite - check if removing the value changes length
                    logger.debug(
                        "json_contains not available for dialect, using json_array_length fallback: %s",
                        e,
                    )
                    result = func.json_array_length(col_expr) > func.coalesce(
                        func.json_array_length(
                            func.json_remove(col_expr, func.json_quote(value_expr))
                        ),
                        0,
                    )
                except Exception as e:
                    # Catch any other unexpected errors and log them before falling back
                    logger.warning(
                        "Unexpected error using json_contains, falling back to json_array_length: %s",
                        e,
                    )
                    result = func.json_array_length(col_expr) > func.coalesce(
                        func.json_array_length(
                            func.json_remove(col_expr, func.json_quote(value_expr))
                        ),
                        0,
                    )
            if expression._alias:
                result = result.label(expression._alias)
            return result

        if op == "array_position":
            col_expr = self._compile(expression.args[0])
            value_expr = self._compile(expression.args[1])
            # Use dialect-specific array position
            # PostgreSQL: array_position(array, value)
            # SQLite: Use json_each with rowid - requires subquery (complex)
            # MySQL: JSON_SEARCH(json_array, 'one', value) - returns path, extract index
            if self.dialect.name in ("postgresql", "duckdb"):
                result = func.array_position(col_expr, value_expr)
            elif self.dialect.name == "sqlite":
                # SQLite: Use json_each to find position
                # We need to use a subquery approach, but for simplicity in compilation,
                # we'll use a CASE expression with json_array_length and iteration
                # This is a simplified approach - a full implementation would use json_each in a subquery
                from sqlalchemy import literal_column, cast

                # For SQLite, we'll use a workaround with json_extract and iteration
                # This is not perfect but works for most cases
                # Full implementation would require a correlated subquery with json_each
                result = literal_column(
                    "NULL"
                )  # Simplified - full implementation requires subquery
            elif self.dialect.name == "mysql":
                # MySQL: JSON_SEARCH returns a path like "$[0]", need to extract index
                # Extract index from JSON path: "$[0]" -> 0
                # Use: CAST(SUBSTRING_INDEX(SUBSTRING_INDEX(JSON_UNQUOTE(JSON_SEARCH(...)), '[', -1), ']', 1) AS UNSIGNED)
                from sqlalchemy import literal_column, cast
                from sqlalchemy.dialects.mysql import INTEGER

                # JSON_SEARCH returns the path, e.g., "$[0]" for index 0
                json_search_result = func.json_search(col_expr, literal("one"), value_expr)

                # Extract index from path: "$[0]" -> "0" -> 0
                # Step 1: JSON_UNQUOTE to remove quotes: "$[0]" -> $[0]
                unquoted = func.json_unquote(json_search_result)
                # Step 2: SUBSTRING_INDEX with '[', -1 to get everything after '[': $[0] -> 0]
                after_bracket = func.substring_index(unquoted, literal("["), literal(-1))
                # Step 3: SUBSTRING_INDEX with ']', 1 to get everything before ']': 0] -> 0
                index_str = func.substring_index(after_bracket, literal("]"), literal(1))
                # Step 4: CAST to UNSIGNED to get integer, subtract 1 for 0-based, add 1 for 1-based result
                # Actually, MySQL array indices are 0-based, but array_position should return 1-based
                # So we add 1: CAST(...) + 1
                result = sa_case(
                    (json_search_result.is_(None), None), else_=cast(index_str, INTEGER) + 1
                )
            else:
                # Generic fallback - use JSON_SEARCH
                result = func.json_search(col_expr, literal("one"), value_expr)
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "array_append":
            col_expr = self._compile(expression.args[0])
            elem_expr = self._compile(expression.args[1])
            if self.dialect.name in ("postgresql", "duckdb"):
                result = func.array_append(col_expr, elem_expr)
            else:
                result = func.json_array_append(col_expr, "$", elem_expr)
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "array_prepend":
            col_expr = self._compile(expression.args[0])
            elem_expr = self._compile(expression.args[1])
            if self.dialect.name in ("postgresql", "duckdb"):
                result = func.array_prepend(elem_expr, col_expr)
            else:
                result = func.json_array_insert(col_expr, "$[0]", elem_expr)
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "array_remove":
            col_expr = self._compile(expression.args[0])
            elem_arg = expression.args[1]
            # For DuckDB, we need the literal value to inline in lambda
            # Check if it's a literal Column or a raw value
            from ..expressions.column import Column as MoltresColumn
            from .builders import format_literal

            if isinstance(elem_arg, MoltresColumn) and elem_arg.op == "literal":
                # It's a literal, get the value
                elem_value = elem_arg.args[0] if elem_arg.args else None
                elem_str = format_literal(elem_value)
            else:
                # Compile it normally and hope it works
                elem_expr = self._compile(elem_arg)
                elem_str = str(elem_expr)

            if self.dialect.name == "postgresql":
                elem_expr = self._compile(elem_arg)
                result = func.array_remove(col_expr, elem_expr)
            elif self.dialect.name == "duckdb":
                # DuckDB uses list_filter to remove elements
                from sqlalchemy import literal_column

                result = literal_column(f"list_filter({col_expr}, x -> x != {elem_str})")
            else:
                from sqlalchemy import literal_column

                elem_expr = self._compile(elem_arg)
                result = literal_column(
                    f"JSON_REMOVE({col_expr}, JSON_SEARCH({col_expr}, 'one', {elem_expr}))"
                )
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "array_distinct":
            col_expr = self._compile(expression.args[0])
            if self.dialect.name == "postgresql":
                from sqlalchemy import literal_column

                result = literal_column(f"ARRAY(SELECT DISTINCT unnest({col_expr}))")
            elif self.dialect.name == "duckdb":
                # DuckDB uses list_distinct
                from sqlalchemy import literal_column

                result = literal_column(f"list_distinct({col_expr})")
            else:
                from sqlalchemy import literal_column

                result = literal_column(f"JSON_ARRAY_DISTINCT({col_expr})")
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "array_sort":
            col_expr = self._compile(expression.args[0])
            if self.dialect.name == "postgresql":
                result = func.array_sort(col_expr)
            elif self.dialect.name == "duckdb":
                # DuckDB uses list_sort
                from sqlalchemy import literal_column

                result = literal_column(f"list_sort({col_expr})")
            else:
                from sqlalchemy import literal_column

                result = literal_column(f"JSON_ARRAY_SORT({col_expr})")
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "array_max":
            col_expr = self._compile(expression.args[0])
            if self.dialect.name == "postgresql":
                result = func.array_max(col_expr)
            elif self.dialect.name == "duckdb":
                # DuckDB uses list_max instead of array_max
                from sqlalchemy import literal_column

                result = literal_column(f"list_max({col_expr})")
            else:
                # SQLite/MySQL: Use json_each to find max value
                # This requires a subquery - simplified implementation
                from sqlalchemy import literal_column

                # For SQLite, use a workaround with json_each
                # SELECT MAX(value) FROM json_each(array_col)
                result = literal_column(f"(SELECT MAX(value) FROM json_each({col_expr}))")
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "array_min":
            col_expr = self._compile(expression.args[0])
            if self.dialect.name == "postgresql":
                result = func.array_min(col_expr)
            elif self.dialect.name == "duckdb":
                # DuckDB uses list_min instead of array_min
                from sqlalchemy import literal_column

                result = literal_column(f"list_min({col_expr})")
            else:
                # SQLite/MySQL: Use json_each to find min value
                from sqlalchemy import literal_column

                # For SQLite, use a workaround with json_each
                result = literal_column(f"(SELECT MIN(value) FROM json_each({col_expr}))")
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "array_sum":
            col_expr = self._compile(expression.args[0])
            if self.dialect.name == "postgresql":
                from sqlalchemy import literal_column

                result = literal_column(f"(SELECT sum(x) FROM unnest({col_expr}) AS x)")
            elif self.dialect.name == "duckdb":
                # DuckDB uses list_sum instead of array_sum
                from sqlalchemy import literal_column

                result = literal_column(f"list_sum({col_expr})")
            else:
                from sqlalchemy import literal_column

                result = literal_column(f"JSON_ARRAY_SUM({col_expr})")
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "rand":
            if len(expression.args) == 0:
                if self.dialect.name == "postgresql":
                    result = func.random()
                elif self.dialect.name == "mysql":
                    result = func.rand()
                else:
                    result = func.random()
            else:
                # seed = expression.args[0]  # Not used in current implementation
                if self.dialect.name == "postgresql":
                    result = func.random()
                elif self.dialect.name == "mysql":
                    result = func.rand()
                else:
                    result = func.random()
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "randn":
            from sqlalchemy import literal_column

            if self.dialect.name == "postgresql":
                result = literal_column("random_normal()")
            elif self.dialect.name == "duckdb":
                # DuckDB doesn't have random_normal, use random() as fallback
                # Note: This is not a true normal distribution
                result = func.random()
            else:
                raise CompilationError(
                    f"randn() is not supported for {self.dialect.name} dialect. "
                    "PostgreSQL may support this with extensions."
                )
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "hash":
            cols = [self._compile(c) for c in expression.args]
            if self.dialect.name == "postgresql":
                if len(cols) == 1:
                    result = func.hashtext(cols[0])
                else:
                    from sqlalchemy import literal_column

                    concat_expr = func.concat(*cols)
                    result = func.hashtext(concat_expr)
            elif self.dialect.name == "mysql":
                if len(cols) == 1:
                    result = func.md5(cols[0])
                else:
                    concat_expr = func.concat(*cols)
                    result = func.md5(concat_expr)
            else:
                from sqlalchemy import literal_column

                if len(cols) == 1:
                    result = literal_column(f"HASH({cols[0]})")
                else:
                    concat_expr = func.concat(*cols)
                    result = literal_column(f"HASH({concat_expr})")
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "md5":
            col_expr = self._compile(expression.args[0])
            if self.dialect.name == "postgresql":
                result = func.md5(col_expr)
            elif self.dialect.name == "mysql":
                result = func.md5(col_expr)
            else:
                # SQLite: md5 is not available by default, requires extension
                # For now, raise an error to indicate limitation
                from sqlalchemy import literal_column

                # Try to use md5 if available, otherwise raise error
                result = literal_column(f"md5({col_expr})")
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "sha1":
            col_expr = self._compile(expression.args[0])
            if self.dialect.name == "postgresql":
                result = func.digest(col_expr, "sha1")
            elif self.dialect.name == "mysql":
                result = func.sha1(col_expr)
            else:
                from sqlalchemy import literal_column

                result = literal_column(f"SHA1({col_expr})")
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "sha2":
            col_expr = self._compile(expression.args[0])
            num_bits_arg = expression.args[1]
            if not isinstance(num_bits_arg, int):
                raise TypeError(
                    f"Expected int for sha2 num_bits, got {type(num_bits_arg).__name__}"
                )
            num_bits: int = num_bits_arg
            if self.dialect.name == "postgresql":
                algo_map = {224: "sha224", 256: "sha256", 384: "sha384", 512: "sha512"}
                algo = algo_map.get(num_bits, "sha256")
                result = func.digest(col_expr, algo)
            elif self.dialect.name == "mysql":
                result = func.sha2(col_expr, num_bits)
            elif self.dialect.name == "duckdb":
                # DuckDB uses sha256, sha384, sha512 functions
                from sqlalchemy import literal_column

                if num_bits == 256:
                    result = func.sha256(col_expr)
                elif num_bits == 384:
                    result = func.sha384(col_expr)
                elif num_bits == 512:
                    result = func.sha512(col_expr)
                else:
                    # Default to sha256
                    result = func.sha256(col_expr)
            else:
                from sqlalchemy import literal_column

                result = literal_column(f"SHA2({col_expr}, {num_bits})")
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "base64":
            col_expr = self._compile(expression.args[0])
            if self.dialect.name == "postgresql":
                result = func.encode(col_expr, "base64")
            elif self.dialect.name == "mysql":
                result = func.to_base64(col_expr)
            elif self.dialect.name == "duckdb":
                # DuckDB requires BLOB type
                from sqlalchemy import cast
                from sqlalchemy import types as sa_types

                blob_expr = cast(col_expr, sa_types.BLOB)
                result = func.base64(blob_expr)
            else:
                from sqlalchemy import literal_column

                result = literal_column(f"BASE64({col_expr})")
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "monotonically_increasing_id":
            # monotonically_increasing_id is implemented as row_number()
            # It always needs an OVER clause, even if empty
            # If used with .over(), the window handling will apply the .over() call
            # Otherwise, we use row_number() OVER () for SQLite compatibility
            result = func.row_number().over()
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "crc32":
            col_expr = self._compile(expression.args[0])
            if self.dialect.name == "mysql":
                result = func.crc32(col_expr)
            else:
                from sqlalchemy import literal_column

                result = literal_column(f"CRC32({col_expr})")
            if expression._alias:
                result = result.label(expression._alias)
            return result
        if op == "soundex":
            col_expr = self._compile(expression.args[0])
            if self.dialect.name == "postgresql":
                result = func.soundex(col_expr)
            elif self.dialect.name == "mysql":
                result = func.soundex(col_expr)
            elif self.dialect.name == "duckdb":
                # DuckDB doesn't support soundex
                raise CompilationError(
                    f"soundex() is not supported for {self.dialect.name} dialect. "
                    "Supported dialects: PostgreSQL, MySQL"
                )
            else:
                from sqlalchemy import literal_column

                result = literal_column(f"SOUNDEX({col_expr})")
            if expression._alias:
                result = result.label(expression._alias)
            return result

        # Try window compiler
        if op.startswith("window_") or op == "window":
            from .expression_compilers.window import compile_window_operation

            result = compile_window_operation(self, op, expression)
            if result is not None:
                return result

        raise CompilationError(
            f"Unsupported expression operation '{op}'. "
            "This may indicate a missing function implementation or an invalid expression.",
            context={
                "operation": op,
                "expression_args": [
                    str(arg) for arg in expression.args[:3]
                ],  # Limit to first 3 args
                "dialect": self.dialect.name,
                "expression_alias": expression._alias,
            },
        )

    def _get_table(self, table_name: str) -> TableClause:
        """Get or create a SQLAlchemy table object for the given table name."""
        from sqlalchemy import table as sa_table

        if table_name not in self._table_cache:
            self._table_cache[table_name] = sa_table(table_name)
        result = self._table_cache[table_name]
        return result  # type: ignore[no-any-return]



"""Shared helper functions for Pandas DataFrame interface operations.

This module contains shared logic used by both PandasDataFrame and AsyncPandasDataFrame
to reduce code duplication and improve maintainability.
"""

from __future__ import annotations

from typing import (
    cast,
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Protocol,
    Sequence,
    TypeVar,
    Union,
)

from ...expressions.column import Column, col

if TYPE_CHECKING:
    from ..core.dataframe import DataFrame
    from ..core.async_dataframe import AsyncDataFrame

# Type variable for generic Pandas DataFrame operations
P = TypeVar("P", bound="PandasDataFrameProtocol")

if TYPE_CHECKING:

    class PandasDataFrameProtocol(Protocol):
        """Protocol defining the interface that Pandas DataFrame classes must implement."""

        _df: Union["DataFrame", "AsyncDataFrame"]
        columns: List[str]

        def _validate_columns_exist(
            self, column_names: Sequence[str], operation: str = "operation"
        ) -> None:
            """Validate that all specified columns exist."""
            ...
else:
    PandasDataFrameProtocol = Any


def build_pandas_query_operation(
    pandas_df: Any,  # PandasDataFrameProtocol - using Any for mypy compatibility
    expr: str,
) -> Union["DataFrame", "AsyncDataFrame"]:
    """Build a query operation for Pandas DataFrame.

    Args:
        pandas_df: Pandas DataFrame instance
        expr: Query string with pandas-style syntax

    Returns:
        Resulting underlying DataFrame
    """
    from ..operations.pandas_operations import parse_query_expression

    # Get available column names for context and validation
    available_columns: Optional[set[str]] = None
    try:
        available_columns = set(pandas_df.columns)
    except Exception:
        pass

    # Parse query string to Column expression
    predicate = parse_query_expression(expr, available_columns, pandas_df._df.plan)

    # Apply filter
    return cast(Union["DataFrame", "AsyncDataFrame"], pandas_df._df.where(predicate))


def build_pandas_select_operation(
    pandas_df: Any,  # PandasDataFrameProtocol - using Any for mypy compatibility
    *columns: Union[str, Column],
) -> Union["DataFrame", "AsyncDataFrame"]:
    """Build a select operation for Pandas DataFrame.

    Args:
        pandas_df: Pandas DataFrame instance
        *columns: Column names or Column expressions to select

    Returns:
        Resulting underlying DataFrame
    """
    # Validate column names if they're strings
    str_columns = [c for c in columns if isinstance(c, str)]
    if str_columns:
        pandas_df._validate_columns_exist(str_columns, "select")

    # Use underlying DataFrame's select
    selected_cols = [col(c) if isinstance(c, str) else c for c in columns]
    return cast(Union["DataFrame", "AsyncDataFrame"], pandas_df._df.select(*selected_cols))


def build_pandas_assign_operation(
    pandas_df: Any,  # PandasDataFrameProtocol - using Any for mypy compatibility
    **kwargs: Union[Column, Any],
) -> Union["DataFrame", "AsyncDataFrame"]:
    """Build an assign operation for Pandas DataFrame.

    Args:
        pandas_df: Pandas DataFrame instance
        **kwargs: Column name = value pairs where value can be a Column expression or literal

    Returns:
        Resulting underlying DataFrame
    """
    result_df = pandas_df._df
    for col_name, value in kwargs.items():
        if isinstance(value, Column):
            result_df = result_df.withColumn(col_name, value)
        else:
            # Literal value
            from ...expressions import lit

            result_df = result_df.withColumn(col_name, lit(value))
    return cast(Union["DataFrame", "AsyncDataFrame"], result_df)


def build_pandas_merge_operation(
    left_pandas_df: Any,  # PandasDataFrameProtocol - using Any for mypy compatibility
    right_pandas_df: Any,  # PandasDataFrameProtocol - using Any for mypy compatibility
    on: Optional[Union[str, Sequence[str]]] = None,
    left_on: Optional[Union[str, Sequence[str]]] = None,
    right_on: Optional[Union[str, Sequence[str]]] = None,
    how: str = "inner",
) -> Union["DataFrame", "AsyncDataFrame"]:
    """Build a merge operation for Pandas DataFrame.

    Args:
        left_pandas_df: Left Pandas DataFrame instance
        right_pandas_df: Right Pandas DataFrame instance
        on: Column name(s) to join on (must exist in both DataFrames)
        left_on: Column name(s) in left DataFrame
        right_on: Column name(s) in right DataFrame
        how: Type of join ('inner', 'left', 'right', 'outer')

    Returns:
        Resulting underlying DataFrame
    """
    from ..operations.pandas_operations import normalize_merge_how, prepare_merge_keys

    # Normalize how parameter
    join_how = normalize_merge_how(how)

    # Determine join keys
    join_on = prepare_merge_keys(
        on=on,
        left_on=left_on,
        right_on=right_on,
        left_columns=left_pandas_df.columns,
        right_columns=right_pandas_df.columns,
        left_validate_fn=left_pandas_df._validate_columns_exist,
        right_validate_fn=right_pandas_df._validate_columns_exist,
    )

    # Perform join
    return cast(
        Union["DataFrame", "AsyncDataFrame"],
        left_pandas_df._df.join(right_pandas_df._df, on=join_on, how=join_how),
    )


def build_pandas_cross_join_operation(
    left_pandas_df: Any,  # PandasDataFrameProtocol - using Any for mypy compatibility
    right_pandas_df: Any,  # PandasDataFrameProtocol - using Any for mypy compatibility
) -> Union["DataFrame", "AsyncDataFrame"]:
    """Build a cross join operation for Pandas DataFrame.

    Args:
        left_pandas_df: Left Pandas DataFrame instance
        right_pandas_df: Right Pandas DataFrame instance

    Returns:
        Resulting underlying DataFrame
    """
    return cast(
        Union["DataFrame", "AsyncDataFrame"], left_pandas_df._df.crossJoin(right_pandas_df._df)
    )


def build_pandas_rename_operation(
    pandas_df: Any,  # PandasDataFrameProtocol - using Any for mypy compatibility
    columns: Dict[str, str],
) -> Union["DataFrame", "AsyncDataFrame"]:
    """Build a rename operation for Pandas DataFrame.

    Args:
        pandas_df: Pandas DataFrame instance
        columns: Dictionary mapping old names to new names

    Returns:
        Resulting underlying DataFrame
    """
    result_df = pandas_df._df
    for old_name, new_name in columns.items():
        result_df = result_df.withColumnRenamed(old_name, new_name)
    return cast(Union["DataFrame", "AsyncDataFrame"], result_df)


def build_pandas_drop_operation(
    pandas_df: Any,  # PandasDataFrameProtocol - using Any for mypy compatibility
    columns: Optional[Union[str, Sequence[str]]] = None,
) -> Union["DataFrame", "AsyncDataFrame"]:
    """Build a drop operation for Pandas DataFrame.

    Args:
        pandas_df: Pandas DataFrame instance
        columns: Column name(s) to drop

    Returns:
        Resulting underlying DataFrame
    """
    if columns is None:
        return cast(Union["DataFrame", "AsyncDataFrame"], pandas_df._df)

    if isinstance(columns, str):
        cols_to_drop = [columns]
    else:
        cols_to_drop = list(columns)

    return cast(Union["DataFrame", "AsyncDataFrame"], pandas_df._df.drop(*cols_to_drop))


def build_pandas_sort_values_operation(
    pandas_df: Any,  # PandasDataFrameProtocol - using Any for mypy compatibility
    by: Union[str, Sequence[str]],
    ascending: Union[bool, Sequence[bool]] = True,
) -> Union["DataFrame", "AsyncDataFrame"]:
    """Build a sort_values operation for Pandas DataFrame.

    Args:
        pandas_df: Pandas DataFrame instance
        by: Column name(s) to sort by
        ascending: Sort order (True for ascending, False for descending)

    Returns:
        Resulting underlying DataFrame
    """
    if isinstance(by, str):
        columns = [by]
        ascending_list = [ascending] if isinstance(ascending, bool) else list(ascending)
    else:
        columns = list(by)
        if isinstance(ascending, bool):
            ascending_list = [ascending] * len(columns)
        else:
            ascending_list = list(ascending)
            if len(ascending_list) != len(columns):
                raise ValueError("ascending must have same length as by")

    # Validate columns exist
    pandas_df._validate_columns_exist(columns, "sort_values")

    # Build order_by list - use Column expressions with .desc() for descending
    order_by_cols = []
    for col_name, asc in zip(columns, ascending_list):
        col_expr = col(col_name)
        if not asc:
            # Descending order
            col_expr = col_expr.desc()
        order_by_cols.append(col_expr)

    return cast(Union["DataFrame", "AsyncDataFrame"], pandas_df._df.order_by(*order_by_cols))


def build_pandas_drop_duplicates_operation(
    pandas_df: Any,  # PandasDataFrameProtocol - using Any for mypy compatibility
    subset: Optional[Union[str, Sequence[str]]] = None,
    keep: str = "first",
) -> Union["DataFrame", "AsyncDataFrame"]:
    """Build a drop_duplicates operation for Pandas DataFrame.

    Args:
        pandas_df: Pandas DataFrame instance
        subset: Column name(s) to consider for duplicates (None means all columns)
        keep: Which duplicate to keep ('first' or 'last')

    Returns:
        Resulting underlying DataFrame
    """
    from ...expressions import functions as F

    if subset is None:
        # Remove duplicates on all columns
        return cast(Union["DataFrame", "AsyncDataFrame"], pandas_df._df.distinct())
    else:
        # Validate subset columns exist
        if isinstance(subset, str):
            subset_cols = [subset]
        else:
            subset_cols = list(subset)

        pandas_df._validate_columns_exist(subset_cols, "drop_duplicates")

        # For subset-based deduplication, use GROUP BY
        grouped = pandas_df._df.group_by(*subset_cols)

        # Get all column names
        all_cols = pandas_df.columns
        other_cols = [col_name for col_name in all_cols if col_name not in subset_cols]

        if not other_cols:
            # If only grouping columns, distinct works fine
            return cast(Union["DataFrame", "AsyncDataFrame"], pandas_df._df.distinct())
        else:
            # Build aggregations for non-grouped columns only
            agg_exprs = []
            for col_name in other_cols:
                if keep == "last":
                    agg_exprs.append(F.max(col(col_name)).alias(col_name))
                else:  # keep == "first"
                    agg_exprs.append(F.min(col(col_name)).alias(col_name))

            return cast(Union["DataFrame", "AsyncDataFrame"], grouped.agg(*agg_exprs))


def build_pandas_limit_operation(
    pandas_df: Any,  # PandasDataFrameProtocol - using Any for mypy compatibility
    n: int,
) -> Union["DataFrame", "AsyncDataFrame"]:
    """Build a limit operation for Pandas DataFrame.

    Args:
        pandas_df: Pandas DataFrame instance
        n: Number of rows to return

    Returns:
        Resulting underlying DataFrame
    """
    return cast(Union["DataFrame", "AsyncDataFrame"], pandas_df._df.limit(n))


def build_pandas_sample_operation(
    pandas_df: Any,  # PandasDataFrameProtocol - using Any for mypy compatibility
    n: Optional[int] = None,
    frac: Optional[float] = None,
    random_state: Optional[int] = None,
) -> Union["DataFrame", "AsyncDataFrame"]:
    """Build a sample operation for Pandas DataFrame.

    Args:
        pandas_df: Pandas DataFrame instance
        n: Number of rows to sample
        frac: Fraction of rows to sample (0.0 to 1.0)
        random_state: Random seed

    Returns:
        Resulting underlying DataFrame
    """
    if n is not None and frac is not None:
        raise ValueError("Cannot specify both 'n' and 'frac'")

    if n is not None:
        # Sample n rows - use fraction=1.0 then limit
        return cast(
            Union["DataFrame", "AsyncDataFrame"],
            pandas_df._df.sample(fraction=1.0, seed=random_state).limit(n),
        )
    elif frac is not None:
        return cast(
            Union["DataFrame", "AsyncDataFrame"],
            pandas_df._df.sample(fraction=frac, seed=random_state),
        )
    else:
        raise ValueError("Must specify either 'n' or 'frac'")


def build_pandas_append_operation(
    left_pandas_df: Any,  # PandasDataFrameProtocol - using Any for mypy compatibility
    right_pandas_df: Any,  # PandasDataFrameProtocol - using Any for mypy compatibility
) -> Union["DataFrame", "AsyncDataFrame"]:
    """Build an append operation for Pandas DataFrame.

    Args:
        left_pandas_df: Left Pandas DataFrame instance
        right_pandas_df: Right Pandas DataFrame instance

    Returns:
        Resulting underlying DataFrame (union all)
    """
    return cast(
        Union["DataFrame", "AsyncDataFrame"], left_pandas_df._df.unionAll(right_pandas_df._df)
    )


def build_pandas_concat_operation(
    pandas_df: Any,  # PandasDataFrameProtocol - using Any for mypy compatibility
    others: Sequence[Any],  # Sequence[PandasDataFrameProtocol] - using Any for mypy compatibility
    axis: Union[int, str] = 0,
) -> Union["DataFrame", "AsyncDataFrame"]:
    """Build a concat operation for Pandas DataFrame.

    Args:
        pandas_df: First Pandas DataFrame instance
        others: Other Pandas DataFrame instances to concatenate
        axis: Concatenation axis (0 for vertical, 1 for horizontal)

    Returns:
        Resulting underlying DataFrame
    """
    if axis == 0 or axis == "index":
        # Vertical concatenation (union all)
        result_df = pandas_df._df
        for other in others:
            result_df = result_df.unionAll(other._df)
        return cast(Union["DataFrame", "AsyncDataFrame"], result_df)
    elif axis == 1 or axis == "columns":
        # Horizontal concatenation (cross join)
        result_df = pandas_df._df
        for other in others:
            result_df = result_df.crossJoin(other._df)
        return cast(Union["DataFrame", "AsyncDataFrame"], result_df)
    else:
        raise ValueError(f"Invalid axis: {axis}. Must be 0, 1, 'index', or 'columns'")


def build_pandas_select_expr_operation(
    pandas_df: Any,  # PandasDataFrameProtocol - using Any for mypy compatibility
    *exprs: str,
) -> Union["DataFrame", "AsyncDataFrame"]:
    """Build a select_expr operation for Pandas DataFrame.

    Args:
        pandas_df: Pandas DataFrame instance
        *exprs: SQL expression strings

    Returns:
        Resulting underlying DataFrame
    """
    return cast(Union["DataFrame", "AsyncDataFrame"], pandas_df._df.selectExpr(*exprs))


def build_pandas_cte_operation(
    pandas_df: Any,  # PandasDataFrameProtocol - using Any for mypy compatibility
    name: str,
) -> Union["DataFrame", "AsyncDataFrame"]:
    """Build a cte operation for Pandas DataFrame.

    Args:
        pandas_df: Pandas DataFrame instance
        name: Name for the CTE

    Returns:
        Resulting underlying DataFrame
    """
    return cast(Union["DataFrame", "AsyncDataFrame"], pandas_df._df.cte(name))


def build_pandas_summary_operation(
    pandas_df: Any,  # PandasDataFrameProtocol - using Any for mypy compatibility
    *statistics: str,
) -> Union["DataFrame", "AsyncDataFrame"]:
    """Build a summary operation for Pandas DataFrame.

    Args:
        pandas_df: Pandas DataFrame instance
        *statistics: Statistics to compute

    Returns:
        Resulting underlying DataFrame
    """
    return cast(Union["DataFrame", "AsyncDataFrame"], pandas_df._df.summary(*statistics))



"""DataFrame schema inspection operations.

This module handles schema inspection operations like columns, schema, dtypes, and printSchema.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, List, Tuple

if TYPE_CHECKING:
    from ...utils.inspector import ColumnInfo
    from ..core.dataframe import DataFrame


class SchemaInspector:
    """Handles schema inspection operations for DataFrames."""

    def __init__(self, df: "DataFrame"):
        """Initialize schema inspector with a DataFrame.

        Args:
            df: The DataFrame to inspect
        """
        self._df = df

    def columns(self) -> List[str]:
        """Return a list of column names in this DataFrame.

        Similar to PySpark's DataFrame.columns property, this extracts column
        names from the logical plan without requiring query execution.

        Returns:
            List of column name strings

        Raises:
            RuntimeError: If column names cannot be determined (e.g., RawSQL without execution)
        """
        return self._df._extract_column_names(self._df.plan)

    def schema(self) -> List["ColumnInfo"]:
        """Return the schema of this DataFrame as a list of ColumnInfo objects.

        Similar to PySpark's DataFrame.schema property, this extracts column
        names and types from the logical plan without requiring query execution.

        Returns:
            List of ColumnInfo objects with column names and types

        Raises:
            RuntimeError: If schema cannot be determined (e.g., RawSQL without execution)
        """
        return self._df._extract_schema_from_plan(self._df.plan)

    def dtypes(self) -> List[Tuple[str, str]]:
        """Return a list of tuples containing column names and their data types.

        Similar to PySpark's DataFrame.dtypes property, this returns a list
        of (column_name, type_name) tuples.

        Returns:
            List of tuples (column_name, type_name)

        Raises:
            RuntimeError: If schema cannot be determined (e.g., RawSQL without execution)
        """
        schema = self.schema()
        return [(col_info.name, col_info.type_name) for col_info in schema]

    def print_schema(self) -> None:
        """Print the schema of this DataFrame in a tree format.

        Similar to PySpark's DataFrame.printSchema() method, this prints
        a formatted representation of the DataFrame's schema.
        """
        schema = self.schema()
        if not schema:
            print("Empty DataFrame")
            return

        print("root")
        for col_info in schema:
            # Format similar to PySpark: |-- column_name: type_name (nullable = true)
            nullable_str = "nullable = true" if col_info.nullable else "nullable = false"
            print(f" |-- {col_info.name}: {col_info.type_name} ({nullable_str})")



"""Async Polars-style GroupBy interface for Moltres."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Dict, Union

from ...expressions.column import Column
from ..groupby.async_groupby import AsyncGroupedDataFrame

if TYPE_CHECKING:
    from ..interfaces.async_polars_dataframe import AsyncPolarsDataFrame


@dataclass(frozen=True)
class AsyncPolarsGroupBy:
    """Async Polars-style GroupBy wrapper around Moltres AsyncGroupedDataFrame.

    Provides Polars-style groupby API with expression-based aggregations.
    """

    _grouped: AsyncGroupedDataFrame

    def agg(self, *exprs: Union[Column, Dict[str, str]]) -> "AsyncPolarsDataFrame":
        """Apply aggregations using Polars-style expressions.

        Args:
            *exprs: :class:`Column` expressions for aggregations, or dictionary mapping column names to function names

        Returns:
            AsyncPolarsDataFrame with aggregated results

        Example:
            >>> await df.group_by('country').agg(col('amount').sum(), col('price').mean())
            >>> await df.group_by('country').agg({"amount": "sum", "price": "avg"})  # Dictionary syntax
        """
        from ..interfaces.async_polars_dataframe import AsyncPolarsDataFrame
        from ..helpers.polars_groupby_helpers import build_polars_groupby_agg_with_dict_handling

        normalized_exprs, _ = build_polars_groupby_agg_with_dict_handling(self, *exprs)
        result_df = self._grouped.agg(*normalized_exprs)
        return AsyncPolarsDataFrame.from_dataframe(result_df)

    def mean(self) -> "AsyncPolarsDataFrame":
        """Mean of all numeric columns in each group.

        Returns:
            AsyncPolarsDataFrame with mean of all numeric columns for each group

        Note:
            This attempts to average all columns. For better control, use agg() with
            specific columns.
        """
        from ..interfaces.async_polars_dataframe import AsyncPolarsDataFrame
        from ...expressions import functions as F
        from ..helpers.polars_groupby_helpers import build_polars_groupby_column_aggregation

        agg_list = build_polars_groupby_column_aggregation(
            self,
            "mean",
            F.avg,
            "_mean",
            "No numeric columns found to average",
        )
        result_df = self._grouped.agg(*agg_list)
        return AsyncPolarsDataFrame.from_dataframe(result_df)

    def sum(self) -> "AsyncPolarsDataFrame":
        """Sum all numeric columns in each group.

        Returns:
            AsyncPolarsDataFrame with sum of all numeric columns for each group
        """
        from ..interfaces.async_polars_dataframe import AsyncPolarsDataFrame
        from ...expressions import functions as F
        from ..helpers.polars_groupby_helpers import build_polars_groupby_column_aggregation

        agg_list = build_polars_groupby_column_aggregation(
            self,
            "sum",
            F.sum,
            "_sum",
            "No numeric columns found to sum",
        )
        result_df = self._grouped.agg(*agg_list)
        return AsyncPolarsDataFrame.from_dataframe(result_df)

    def min(self) -> "AsyncPolarsDataFrame":
        """Minimum value of all columns in each group."""
        from ..interfaces.async_polars_dataframe import AsyncPolarsDataFrame
        from ...expressions import functions as F
        from ..helpers.polars_groupby_helpers import build_polars_groupby_column_aggregation

        agg_list = build_polars_groupby_column_aggregation(
            self,
            "min",
            F.min,
            "_min",
            "No columns found for min()",
        )
        result_df = self._grouped.agg(*agg_list)
        return AsyncPolarsDataFrame.from_dataframe(result_df)

    def max(self) -> "AsyncPolarsDataFrame":
        """Maximum value of all columns in each group."""
        from ..interfaces.async_polars_dataframe import AsyncPolarsDataFrame
        from ...expressions import functions as F
        from ..helpers.polars_groupby_helpers import build_polars_groupby_column_aggregation

        agg_list = build_polars_groupby_column_aggregation(
            self,
            "max",
            F.max,
            "_max",
            "No columns found for max()",
        )
        result_df = self._grouped.agg(*agg_list)
        return AsyncPolarsDataFrame.from_dataframe(result_df)

    def count(self) -> "AsyncPolarsDataFrame":
        """Count rows in each group."""
        from ...expressions.functions import count
        from ..interfaces.async_polars_dataframe import AsyncPolarsDataFrame

        result_df = self._grouped.agg(count("*").alias("count"))
        return AsyncPolarsDataFrame.from_dataframe(result_df)

    def std(self) -> "AsyncPolarsDataFrame":
        """Standard deviation of all numeric columns in each group."""
        from ..interfaces.async_polars_dataframe import AsyncPolarsDataFrame
        from ...expressions import functions as F
        from ..helpers.polars_groupby_helpers import build_polars_groupby_column_aggregation

        agg_list = build_polars_groupby_column_aggregation(
            self,
            "std",
            F.stddev,
            "_std",
            "No numeric columns found for std()",
        )
        result_df = self._grouped.agg(*agg_list)
        return AsyncPolarsDataFrame.from_dataframe(result_df)

    def var(self) -> "AsyncPolarsDataFrame":
        """Variance of all numeric columns in each group."""
        from ..interfaces.async_polars_dataframe import AsyncPolarsDataFrame
        from ...expressions import functions as F
        from ..helpers.polars_groupby_helpers import build_polars_groupby_column_aggregation

        agg_list = build_polars_groupby_column_aggregation(
            self,
            "var",
            F.variance,
            "_var",
            "No numeric columns found for var()",
        )
        result_df = self._grouped.agg(*agg_list)
        return AsyncPolarsDataFrame.from_dataframe(result_df)

    def first(self) -> "AsyncPolarsDataFrame":
        """Get first value of each column in each group."""
        from ..interfaces.async_polars_dataframe import AsyncPolarsDataFrame
        from ...expressions import functions as F
        from ..helpers.polars_groupby_helpers import build_polars_groupby_column_aggregation

        # Use MIN as proxy for first (works if data is ordered)
        agg_list = build_polars_groupby_column_aggregation(
            self,
            "first",
            F.min,
            "_first",
            "No columns found for first()",
        )
        result_df = self._grouped.agg(*agg_list)
        return AsyncPolarsDataFrame.from_dataframe(result_df)

    def last(self) -> "AsyncPolarsDataFrame":
        """Get last value of each column in each group."""
        from ..interfaces.async_polars_dataframe import AsyncPolarsDataFrame
        from ...expressions import functions as F
        from ..helpers.polars_groupby_helpers import build_polars_groupby_column_aggregation

        # Use MAX as proxy for last (works if data is ordered)
        agg_list = build_polars_groupby_column_aggregation(
            self,
            "last",
            F.max,
            "_last",
            "No columns found for last()",
        )
        result_df = self._grouped.agg(*agg_list)
        return AsyncPolarsDataFrame.from_dataframe(result_df)

    def n_unique(self) -> "AsyncPolarsDataFrame":
        """Count distinct values for all columns in each group."""
        from ..interfaces.async_polars_dataframe import AsyncPolarsDataFrame
        from ...expressions import functions as F
        from ..helpers.polars_groupby_helpers import build_polars_groupby_column_aggregation

        agg_list = build_polars_groupby_column_aggregation(
            self,
            "n_unique",
            F.count_distinct,
            "_n_unique",
            "No columns found for n_unique()",
        )
        result_df = self._grouped.agg(*agg_list)
        return AsyncPolarsDataFrame.from_dataframe(result_df)
